------------------------------------------------------------------------
-- This software is provided as is in the hope that it might be found
-- useful.
-- You may use and modify it freely provided you keep this copyright
-- notice unchanged and mark modifications appropriately.
--
-- Bug reports and proposals for improvements are welcome. Please send
-- them to the eMail address below.
--
-- Christoph Karl Walter Grein
-- Hauptstr. 42
-- D-86926 Greifenberg
-- Germany
--
-- eMail: Christ-Usch.Grein@T-Online.de
--
-- Copyright (c) 1998 Christoph Karl Walter Grein
------------------------------------------------------------------------

with Ada.Command_Line;
with Ada.Text_IO;
use  Ada.Text_IO;

with Lexer.Read_File;  pragma Elaborate_All (Lexer);

with HTML, RTF;

procedure Colorize is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   2.3
  -- Date      27 October 1998
  --====================================================================
  -- Colorize an Ada file given as command argument in HyperText Markup
  -- Language (HTML) or Rich Text Format (RTF).
  -- Output it with extension ".html" respectively ".rtf" appended.
  --
  -- Syntax:
  --   Colorize {-option value} file_name
  -- where option is one of {text, file, tab} and value depends on the
  -- option.
  --
  -- Value for text option:
  --   None : Just reconstruct the code as is without adding HTML or RTF
  --          control strings.
  --   Color: Display tokens in different colors or
  --   Bold :   font styles according to some classification.
  --
  -- Value for file option:
  --   HTML: Create a HTML file respectively
  --   RTF :   a RTF file.
  --
  -- Value for tab option:
  --   A positive integer denoting the number of spaces a horizontal
  --   tabulation occupies.
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    0.0  30.05.1998 Just reconstruct an Ada file
  --  C.G.    0.1  11.06.1998 Lexer interface was changed
  --  C.G.    1.0  30.06.1998 Color option
  --  C.G.    2.0  07.07.1998 HTML file, size of tab
  --  C.G.    2.1  30.07.1998 Adapt to Java enabled Lexer
  --  C.G.    2.2  03.10.1998 Separate color for documentation tags
  --  C.G.    2.3  27.10.1998 Use Lexer.Read_File; new colour names
  --====================================================================

  -- Check for correct usage

  Usage: exception;

  Tab_Size: Positive := 8;

  type Text_Option is (None, Color, Bold);
  Text_Option_Selected: Text_Option := Bold;

  type File_Option is (HTML, RTF);
  File_Option_Selected: File_Option := HTML;

  Extension: constant array (File_Option) of Lexer.String_Pointer :=
    (HTML => new String'(".html"),
     RTF  => new String'(".rtf"));

  procedure Check_Arguments is
    type Option is (Text, File, Tab);
    File_Option_Set, Text_Option_Set, Tab_Option_Set, File_Given: Boolean := False;
    Arg: Positive := 1;
    use Ada.Command_Line;
  begin
    if Argument_Count = 0 then
      raise Usage;
    end if;
    while Arg <= Argument_Count loop
      if Argument (Arg)(1) = '-' then
        case Option'Value (Argument (Arg) (2 .. Argument (Arg)'Length)) is
          when Text => if Text_Option_Set or File_Given then
                         raise Usage;
                       end if;
                       Arg := Arg + 1;
                       Text_Option_Set      := True;
                       Text_Option_Selected := Text_Option'Value (Argument (Arg));
          when File => if File_Option_Set or File_Given then
                         raise Usage;
                       end if;
                       Arg := Arg + 1;
                       File_Option_Set      := True;
                       File_Option_Selected := File_Option'Value (Argument (Arg));
          when Tab  => if Tab_Option_Set or File_Given then
                         raise Usage;
                       end if;
                       Arg := Arg + 1;
                       Tab_Option_Set := True;
                       Tab_Size       := Positive'Value (Argument (Arg));
        end case;
      else
        if File_Given then
          raise Usage;
        end if;
        File_Given := True;
      end if;
      Arg := Arg + 1;
    end loop;
    if not File_Given then
      raise Usage;
    end if;
    -- No errors on command line and file name is last parameter.
  exception
    when Constraint_Error =>
      raise Usage;
  end Check_Arguments;

  -----------------------------------------------------------------------
  -- Original file

  procedure Read_Original is
    use Ada.Command_Line;
  begin
    -- This is a storage leak, since the file pointer is not deallocated.
    -- The program is however considered to be used stand-alone.
    Lexer.Initialize (Lexer.Read_File (Argument (Argument_Count)),
                      Tab_Size);
  end Read_Original;

  -----------------------------------------------------------------------
  -- Token processing

  type Token_Class is
    (Structure, Reserved, Literal, Characters, Comment, Documentation_Tag, Bad, Other_Class);

  function Classify (Item: Lexer.Token; Tag: in Boolean) return Token_Class is
    use Lexer;
  begin
    case Lexer.Name (Item) is
      when LeftBrace_J   |
           RightBrace_J  => return Structure;
      when Reserved_Word => return Reserved;
      when Lexer.Literal => if Lexer.Name (Item) >= Character_AJ then
                              return Characters;
                            else
                              return Literal;
                            end if;
      when Lexer.Comment => if Tag then
                              return Documentation_Tag;
                            else
                              return Comment;
                            end if;
      when Bad_Token_AJ  => return Bad;
      when others        => return Other_Class;
    end case;
  end Classify;

  type Token_Style is
    (Plain, Italic, Bold, Bold_Italic);

  function Stylize (Item: Lexer.Token; Tag: in Boolean) return Token_Style is
    use Lexer;
  begin
    case Classify (Item, Tag) is
      when Structure         |
           Reserved          => return Bold;
      when Comment           => return Italic;
      when Documentation_Tag |
           Bad               => return Bold_Italic;
      when others            => return Plain;
    end case;
  end Stylize;

  -----------------------------------------------------------------------
  -- Output file

  Colorized: Ada.Text_IO.File_Type;
  -- Log: Ada.Text_IO.File_Type;
  Current_Line, Current_Col: Positive := 1;

  Token: Lexer.Token;
  use type Lexer.Token_Name;
  Current_Class: Token_Class := Other_Class;
  Current_Style: Token_Style := Plain;

  HTML_Color: constant array (Token_Class) of Lexer.String_Pointer :=
    (Structure         => new String'(Standard.HTML.Fuchsia),
     Reserved          => new String'(Standard.HTML.Blue),
     Literal           => new String'(Standard.HTML.Purple),
     Characters        => new String'(Standard.HTML.Teal),
     Comment           => new String'(Standard.HTML.Green),
     Documentation_Tag => new String'(Standard.HTML.Olive),
     Bad               => new String'(Standard.HTML.Red),
     Other_Class       => new String'(Standard.HTML.No_Color));
  HTML_Style: constant array (Token_Style) of Lexer.String_Pointer :=
    (Italic            => new String'(Standard.HTML.Italic),
     Bold              => new String'(Standard.HTML.Bold),
     Bold_Italic       => new String'(Standard.HTML.Bold_Italic),
     Plain             => new String'(""));
  HTML_Un_Style: constant array (Token_Style) of Lexer.String_Pointer :=
    (Italic            => new String'(Standard.HTML.No_Italic),
     Bold              => new String'(Standard.HTML.No_Bold),
     Bold_Italic       => new String'(Standard.HTML.No_Bold_Italic),
     Plain             => new String'(""));
  RTF_Color: constant array (Token_Class) of Lexer.String_Pointer :=
    (Structure         => new String'(Standard.RTF.Fuchsia),
     Reserved          => new String'(Standard.RTF.Blue),
     Literal           => new String'(Standard.RTF.Purple),
     Characters        => new String'(Standard.RTF.Teal),
     Comment           => new String'(Standard.RTF.Green),
     Documentation_Tag => new String'(Standard.RTF.Olive),
     Bad               => new String'(Standard.RTF.Red),
     Other_Class       => new String'(Standard.RTF.No_Color));
  RTF_Style: constant array (Token_Style) of Lexer.String_Pointer :=
    (Bold              => new String'(Standard.RTF.Bold),
     Italic            => new String'(Standard.RTF.Italic),
     Bold_Italic       => new String'(Standard.RTF.Bold_Italic),
     Plain             => new String'(Standard.RTF.Plain));

  procedure Write_Header is
  begin
    if Text_Option_Selected /= None then
      case File_Option_Selected is
        when HTML => Put      (Colorized, Standard.HTML.Header
                                            (1 .. Standard.HTML.Title_Index));
                     Put      (Colorized, Name (Colorized));
                     Put_Line (Colorized, Standard.HTML.Header
                                            (Standard.HTML.Title_Index + 1 .. Standard.HTML.Header'Last));
        when RTF  => Put_Line (Colorized, Standard.RTF .Header);
      end case;
    end if;
  end Write_Header;

  procedure Write_Footer is
  begin
    if Text_Option_Selected /= None then
      case File_Option_Selected is
        when HTML => Put_Line (Colorized, Standard.HTML.Footer);
        when RTF  => Put_Line (Colorized, Standard.RTF .Footer);
      end case;
    end if;
  end Write_Footer;

  procedure Add_Spacing is
  begin
    for Line in Current_Line + 1 .. Lexer.Line (Token) loop
      New_Line (Colorized);
      if Text_Option_Selected /= None and File_Option_Selected = RTF then
        Put (Colorized, Standard.RTF.BOL);
      end if;
      Current_Line := Current_Line + 1;
      Current_Col  := 1;
    end loop;
    for Col in Current_Col + 1 .. Lexer.Col (Token) loop
      Put (Colorized, ' ');
      Current_Col := Current_Col + 1;
    end loop;
  end Add_Spacing;

  procedure Add (New_Class: in Token_Class) is
  begin
    if New_Class /= Current_Class then
      case File_Option_Selected is
        when HTML => if Current_Class /= Other_Class and New_Class /= Other_Class then
                       Put (Colorized, HTML_Color (Other_Class).all);
                     end if;
                     Put (Colorized, HTML_Color (New_Class).all);
        when RTF  => Put (Colorized, RTF_Color (New_Class).all);
      end case;
    end if;
    Current_Class := New_Class;
  end Add;

  procedure Add (New_Style: in Token_Style) is
  begin
    if New_Style /= Current_Style then
      case File_Option_Selected is
        when HTML => Put (Colorized, HTML_Un_Style (Current_Style).all);
                     Put (Colorized, HTML_Style    (New_Style)    .all);
        when RTF  => if Current_Style /= Plain and New_Style /= Plain then
                       Put (Colorized, RTF_Style (Plain).all);
                     end if;
                     Put (Colorized, RTF_Style (New_Style).all);
      end case;
    end if;
    Current_Style := New_Style;
  end Add;

  procedure Add_Text_Modifier (Tag: in Boolean := False) is
  begin
    case Text_Option_Selected is
      when None  => null;
      when Color => Add (Classify (Token, Tag));
      when Bold  => Add (Stylize  (Token, Tag));
    end case;
  end Add_Text_Modifier;

  procedure Add (Text: in String) is
    -- Count only proper text (do not count modifiers).
    -- Note that strings with a tabulation inside (illegal) lead to
    -- incorrect column counting upon reconstruction. It is not considered
    -- worthwhile to correct this behaviour since it is illegal anyway.
    Found: Boolean;
  begin
    for I in Text'Range loop
      if Text_Option_Selected /= None then
        Found := False;
        case File_Option_Selected is
          when HTML => Special_HTML_Treatment:
                       for S in Standard.HTML.Replace'Range loop
                         if Text (I) = Standard.HTML.Replace (S).Source then
                           Put (Colorized, Standard.HTML.Replace (S).Target.all);
                           Found := True;
                           exit Special_HTML_Treatment;
                         end if;
                       end loop Special_HTML_Treatment;
          when RTF  => Special_RTF_Treatment:
                       for S in Standard.RTF.Replace'Range loop
                         if Text (I) = Standard.RTF.Replace (S).Source then
                           Put (Colorized, Standard.RTF.Replace (S).Target.all);
                           Found := True;
                           exit Special_RTF_Treatment;
                         end if;
                       end loop Special_RTF_Treatment;
        end case;
        if not Found then
          Put (Colorized, Text (I));
        end if;
      else
        Put (Colorized, Text (I));
      end if;
    end loop;
    Current_Col := Current_Col + Text'Length;
  end Add;

  -- procedure Write_Log is
  -- begin
  --   Put_Line (Log, Lexer.Token_Name'Image (Lexer.Name (Token)) &
  --                  Integer'Image (Lexer.First (Token)) &
  --                  Integer'Image (Lexer.Last  (Token)) &
  --                  Integer'Image (Lexer.Line  (Token)) &
  --                  Integer'Image (Lexer.Col   (Token)) &
  --                  Integer'Image (Lexer.Image (Token)'Length));
  --   for C in Lexer.Image (Token)'Range loop
  --     Put (Log, Character'Image (Lexer.Image (Token) (C)));
  --   end loop;
  --   if Lexer.Name (Token) = Lexer.Bad_Token_AJ then
  --     Put (Log, ' ' & Lexer.Token_Error'Image (Lexer.Error (Token)));
  --   end if;
  --   New_Line (Log);
  -- end Write_Log;

  use type Lexer.Documentation_Tag;

begin

  Check_Arguments;

  Read_Original;

  Create (Name => Ada.Command_Line.Argument (Ada.Command_Line.Argument_Count) &
                  Extension (File_Option_Selected).all,
          File => Colorized,
          Mode => Out_File);
       -- Form => "Other => None, Group => None, Owner => Read_Write, " &
       --         "Replace => True");  -- for Rational Apex on Sun

  -- Create (Name => Ada.Command_Line.Argument (Ada.Command_Line.Argument_Count) & ".log",
  --         File => Log,
  --         Mode => Out_File);

  Write_Header;

  while not Lexer.End_of_Input loop

    Token := Lexer.Get_Token;

    -- Write_Log;

    Add_Spacing;

    if Lexer.Name (Token) in Lexer.Documentation and then
       Lexer.Tag  (Token) /= Lexer.No_Tag then
      declare
        Token_Image: constant String := Lexer.Image (Token);
        Tag_Image  : constant String := Lexer.Image (Lexer.Tag (Token));
        Tag_Begin, After_Tag: Natural;
      begin
        Tag_Begin := Lexer.Tag_Pos (Token);
        After_Tag := Tag_Begin + Tag_Image'Length;
        Add_Text_Modifier (Tag => False);  -- part before tag
        Add (Token_Image (Token_Image'First .. Tag_Begin - 1));
        Add_Text_Modifier (Tag => True);   -- the tag itself
        Add (Tag_Image);
        Add_Text_Modifier (Tag => False);  -- part after tag
        Add (Token_Image (After_Tag .. Token_Image'Last));
      end;
    else
      Add_Text_Modifier;
      Add (Lexer.Image (Token));
    end if;

  end loop;

  Write_Footer;

  Close (Colorized);
  -- Close (Log);

exception

  when Name_Error =>
    Ada.Text_IO.Put_Line ("Cannot find file.");
  when Usage =>
    Put_Line ("Usage:");
    Put_Line (Ada.Command_Line.Command_Name & " {-option value} file_name");
    Put_Line ("option is one of {text, file, tab} and value depends on the option:");
    Put_Line ("Value for text option:");
    Put_Line ("  None : Just reconstruct the code as is without adding HTML or RTF control strings.");
    Put_Line ("  Color: Display tokens in different colors or");
    Put_Line ("  Bold :   font styles according to some classification.");
    Put_Line ("Default is bold");
    Put_Line ("Value for file option:");
    Put_Line ("  HTML: Create a HTML file respectively");
    Put_Line ("  RTF :   a RTF file.");
    Put_Line ("Default is HTML");
    Put_Line ("Value for tab option:");
    Put_Line ("  A positive integer denoting the number of spaces a horizontal");
    Put_Line ("  tabulation occupies.");
    Put_Line ("Default is 8");

end Colorize;
