------------------------------------------------------------------------
-- This software is provided as is in the hope that it might be found
-- useful.
-- You may use and modify it freely provided you keep this copyright
-- notice unchanged and mark modifications appropriately.
--
-- Bug reports and proposals for improvements are welcome. Please send
-- them to the eMail address below.
--
-- Christoph Karl Walter Grein
-- Hauptstr. 42
-- D-86926 Greifenberg
-- Germany
--
-- eMail: Christ-Usch.Grein@T-Online.de
--
-- Copyright (c) 1998 Christoph Karl Walter Grein
------------------------------------------------------------------------

with Ada.Characters.Handling,
     Ada.Characters.Latin_1,
     Ada.Strings.Fixed;

package body Lexer is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   3.5
  -- Date      21 November 1998
  --====================================================================
  -- Created following the free software LEXER available with AdaGIDE,
  -- Ada Gnat Integrated Development Environment by Dr. Martin C.
  -- Carlisle.
  --
  -- The lexer processes an input string (a stream of characters)
  -- without line and column information.
  -- Get_Next_Char is the basic operation for reading the input stream.
  -- It performs line and column counting (we call this the file
  -- position) with each EOL encountered, where EOL is defined as the
  -- character sequence {CR}LF. [Windows95 has CR, Unix has not.]
  -- When the end of the stream is passed, EOT is returned, which will
  -- be transformed into the End_Error exception when Get tries to
  -- construct the next token.
  --
  -- An Ada source begins with one of the following items:
  --  - a comment
  --  - a context clause
  --  - a library item or subunit
  --  - a pragma
  -- A Java source begins with one of the following items:
  --  - a comment
  --  - a package statement
  --  - an import statement
  --  - a type declaration
  -- So the lexer can decide by itself which kind of source it is
  -- analysing by looking at the first few tokens, which must be:
  -- For Ada:
  --   Either "--" or one of the reserved words
  --   {function, generic, package, pragma, private, procedure,
  --    separate, with}.
  -- For Java:
  --   Either "//" or "/*" or one of the reserved words
  --   {abstract, class, final, import, interface, package, public}.
  -- Note that only for "package" there is an ambiguity, which can be
  -- resolved by looking which of "is" or ";" appears first. In the
  -- former case, it is an Ada file, in the latter a Java file. This
  -- implies that an illegal Ada or Java unit might not be recognized
  -- as such.
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  23.06.1998
  --  C.G.    1.1  26.06.1998 Bug fixes for Unix (no CR)
  --  C.G.    2.0  01.07.1998 Make Token controlled, more bug fixes
  --  C.G.    2.1  06.07.1998 Size of Tab
  --  C.G.    3.0  30.07.1998 Also lex Java; bug fix in Get_Line.
  --                          [Does not yet lex:
  --                          - Unicode escapes (JRM 3.3);
  --                          - character (JRM 3.10.4) and string (JRM
  --                            3.10.5) literals with escape sequences;
  --                          - based integers (octal and hexadecimal);
  --                          - lazy reals (e.g. .1 instead of 0.1).]
  --  C.G.    3.1  05.08.1998 Java language recognition corrected
  --                          (modifiers according to JRM 8.1.2, 9.1.2);
  --                          simple character escapes ('\n')
  --  C.G.    3.2  18.09.1998 Java octal escape characters and based
  --                          integers (octal and hexadecimal);
  --                          procedure Put_Back_Last_Character is new
  --  C.G.    3.3  22.09.1998 Replace Ampersand_AJ and Vertical_Bar_AJ
  --  C.G.    3.4  03.10.1998 Bug fix for comment continuation lines;
  --                          added functions Is_Operator, for
  --                          documentation tags Image and Tag_Pos
  --  C.G.    3.5  21.11.1998 Bug fix in Reset; simplifications
  --====================================================================

  use Ada;  -- to avoid name clash with language type below

  type Language is (Unknown, Ada, Java);

  Program_Language: Language;

  -- Association of reserved words to language (for unknown, only the
  -- possible initial ones are taken)

  Is_Reserved: constant array (Reserved_Word, Language) of Boolean :=
    (Abort_A        => (Unknown => False, Ada => True , Java => False),
     Abs_A          => (Unknown => False, Ada => True , Java => False),
     Abstract_AJ    => (Unknown => True , Ada => True , Java => True ),
     Accept_A       => (Unknown => False, Ada => True , Java => False),
     Access_A       => (Unknown => False, Ada => True , Java => False),
     Aliased_A      => (Unknown => False, Ada => True , Java => False),
     All_A          => (Unknown => False, Ada => True , Java => False),
     And_A          => (Unknown => False, Ada => True , Java => False),
     Array_A        => (Unknown => False, Ada => True , Java => False),
     At_A           => (Unknown => False, Ada => True , Java => False),
     Begin_A        => (Unknown => False, Ada => True , Java => False),
     Body_A         => (Unknown => False, Ada => True , Java => False),
     Boolean_J      => (Unknown => False, Ada => False, Java => True ),
     Break_J        => (Unknown => False, Ada => False, Java => True ),
     Byte_J         => (Unknown => False, Ada => False, Java => True ),
     Case_AJ        => (Unknown => False, Ada => True , Java => True ),
     Catch_J        => (Unknown => False, Ada => False, Java => True ),
     Char_J         => (Unknown => False, Ada => False, Java => True ),
     Class_J        => (Unknown => True , Ada => False, Java => True ),
     Const_J        => (Unknown => False, Ada => False, Java => True ),
     Constant_A     => (Unknown => False, Ada => True , Java => False),
     Continue_J     => (Unknown => False, Ada => False, Java => True ),
     Declare_A      => (Unknown => False, Ada => True , Java => False),
     Default_J      => (Unknown => False, Ada => False, Java => True ),
     Delay_A        => (Unknown => False, Ada => True , Java => False),
     Delta_A        => (Unknown => False, Ada => True , Java => False),
     Digits_A       => (Unknown => False, Ada => True , Java => False),
     Do_AJ          => (Unknown => False, Ada => True , Java => True ),
     Double_J       => (Unknown => False, Ada => False, Java => True ),
     Else_AJ        => (Unknown => False, Ada => True , Java => True ),
     Elsif_A        => (Unknown => False, Ada => True , Java => False),
     End_A          => (Unknown => False, Ada => True , Java => False),
     Entry_A        => (Unknown => False, Ada => True , Java => False),
     Exception_A    => (Unknown => False, Ada => True , Java => False),
     Exit_A         => (Unknown => False, Ada => True , Java => False),
     Extends_J      => (Unknown => False, Ada => False, Java => True ),
     Final_J        => (Unknown => True , Ada => False, Java => True ),
     Finally_J      => (Unknown => False, Ada => False, Java => True ),
     Float_J        => (Unknown => False, Ada => False, Java => True ),
     For_AJ         => (Unknown => False, Ada => True , Java => True ),
     Function_A     => (Unknown => True , Ada => True , Java => False),
     Generic_A      => (Unknown => True , Ada => True , Java => False),
     Goto_AJ        => (Unknown => False, Ada => True , Java => True ),
     If_AJ          => (Unknown => False, Ada => True , Java => True ),
     Implements_J   => (Unknown => False, Ada => False, Java => True ),
     Import_J       => (Unknown => True , Ada => False, Java => True ),
     In_A           => (Unknown => False, Ada => True , Java => False),
     InstanceOf_J   => (Unknown => False, Ada => False, Java => True ),
     Int_J          => (Unknown => False, Ada => False, Java => True ),
     Interface_J    => (Unknown => True , Ada => False, Java => True ),
     Is_A           => (Unknown => True , Ada => True , Java => False),
     Limited_A      => (Unknown => False, Ada => True , Java => False),
     Long_J         => (Unknown => False, Ada => False, Java => True ),
     Loop_A         => (Unknown => False, Ada => True , Java => False),
     Mod_A          => (Unknown => False, Ada => True , Java => False),
     Native_J       => (Unknown => False, Ada => False, Java => True ),
     New_AJ         => (Unknown => False, Ada => True , Java => True ),
     Not_A          => (Unknown => False, Ada => True , Java => False),
     Null_A         => (Unknown => False, Ada => True , Java => False),
     Of_A           => (Unknown => False, Ada => True , Java => False),
     Or_A           => (Unknown => False, Ada => True , Java => False),
     Others_A       => (Unknown => False, Ada => True , Java => False),
     Out_A          => (Unknown => False, Ada => True , Java => False),
     Package_AJ     => (Unknown => True , Ada => True , Java => True ),
     Pragma_A       => (Unknown => True , Ada => True , Java => False),
     Private_AJ     => (Unknown => True , Ada => True , Java => True ),
     Procedure_A    => (Unknown => True , Ada => True , Java => False),
     Protected_AJ   => (Unknown => False, Ada => True , Java => True ),
     Public_J       => (Unknown => True , Ada => False, Java => True ),
     Raise_A        => (Unknown => False, Ada => True , Java => False),
     Range_A        => (Unknown => False, Ada => True , Java => False),
     Record_A       => (Unknown => False, Ada => True , Java => False),
     Rem_A          => (Unknown => False, Ada => True , Java => False),
     Renames_A      => (Unknown => False, Ada => True , Java => False),
     Requeue_A      => (Unknown => False, Ada => True , Java => False),
     Return_AJ      => (Unknown => False, Ada => True , Java => True ),
     Reverse_A      => (Unknown => False, Ada => True , Java => False),
     Select_A       => (Unknown => False, Ada => True , Java => False),
     Separate_A     => (Unknown => True , Ada => True , Java => False),
     Short_J        => (Unknown => False, Ada => False, Java => True ),
     Static_J       => (Unknown => False, Ada => False, Java => True ),
     Subtype_A      => (Unknown => False, Ada => True , Java => False),
     Super_J        => (Unknown => False, Ada => False, Java => True ),
     Switch_J       => (Unknown => False, Ada => False, Java => True ),
     Synchronized_J => (Unknown => False, Ada => False, Java => True ),
     Tagged_A       => (Unknown => False, Ada => True , Java => False),
     Task_A         => (Unknown => False, Ada => True , Java => False),
     Terminate_A    => (Unknown => False, Ada => True , Java => False),
     Then_A         => (Unknown => False, Ada => True , Java => False),
     This_J         => (Unknown => False, Ada => False, Java => True ),
     Throw_J        => (Unknown => False, Ada => False, Java => True ),
     Throws_J       => (Unknown => False, Ada => False, Java => True ),
     Transient_J    => (Unknown => False, Ada => False, Java => True ),
     Try_J          => (Unknown => False, Ada => False, Java => True ),
     Type_A         => (Unknown => False, Ada => True , Java => False),
     Until_A        => (Unknown => False, Ada => True , Java => False),
     Use_A          => (Unknown => False, Ada => True , Java => False),
     Void_J         => (Unknown => False, Ada => False, Java => True ),
     Volatile_J     => (Unknown => False, Ada => False, Java => True ),
     When_A         => (Unknown => False, Ada => True , Java => False),
     While_AJ       => (Unknown => False, Ada => True , Java => True ),
     With_A         => (Unknown => True , Ada => True , Java => False),
     Xor_A          => (Unknown => False, Ada => True , Java => False));

  -- Images

  Reserved_Word_Image: constant array (Reserved_Word) of String_Pointer :=
    (Abort_A        => new String'("abort"),
     Abs_A          => new String'("abs"),
     Abstract_AJ    => new String'("abstract"),
     Accept_A       => new String'("accept"),
     Access_A       => new String'("access"),
     Aliased_A      => new String'("aliased"),
     All_A          => new String'("all"),
     And_A          => new String'("and"),
     Array_A        => new String'("array"),
     At_A           => new String'("at"),
     Begin_A        => new String'("begin"),
     Body_A         => new String'("body"),
     Boolean_J      => new String'("boolean"),
     Break_J        => new String'("break"),
     Byte_J         => new String'("byte"),
     Case_AJ        => new String'("case"),
     Catch_J        => new String'("catch"),
     Char_J         => new String'("char"),
     Class_J        => new String'("class"),
     Const_J        => new String'("const"),
     Constant_A     => new String'("constant"),
     Continue_J     => new String'("continue"),
     Declare_A      => new String'("declare"),
     Default_J      => new String'("default"),
     Delay_A        => new String'("delay"),
     Delta_A        => new String'("delta"),
     Digits_A       => new String'("digits"),
     Do_AJ          => new String'("do"),
     Double_J       => new String'("double"),
     Else_AJ        => new String'("else"),
     Elsif_A        => new String'("elsif"),
     End_A          => new String'("end"),
     Entry_A        => new String'("entry"),
     Exception_A    => new String'("exception"),
     Exit_A         => new String'("exit"),
     Extends_J      => new String'("extends"),
     Final_J        => new String'("final"),
     Finally_J      => new String'("finally"),
     Float_J        => new String'("float"), 
     For_AJ         => new String'("for"),
     Function_A     => new String'("function"),
     Generic_A      => new String'("generic"),
     Goto_AJ        => new String'("goto"),
     If_AJ          => new String'("if"),
     Implements_J   => new String'("implements"),
     Import_J       => new String'("import"),
     In_A           => new String'("in"),
     InstanceOf_J   => new String'("instanceof"),
     Int_J          => new String'("int"),
     Interface_J    => new String'("interface"),
     Is_A           => new String'("is"),
     Limited_A      => new String'("limited"),
     Long_J         => new String'("long"),
     Loop_A         => new String'("loop"),
     Mod_A          => new String'("mod"),
     Native_J       => new String'("native"),
     New_AJ         => new String'("new"),
     Not_A          => new String'("not"),
     Null_A         => new String'("null"),
     Of_A           => new String'("of"),
     Or_A           => new String'("or"),
     Others_A       => new String'("others"),
     Out_A          => new String'("out"),
     Package_AJ     => new String'("package"),
     Pragma_A       => new String'("pragma"),
     Private_AJ     => new String'("private"),
     Procedure_A    => new String'("procedure"),
     Protected_AJ   => new String'("protected"),
     Public_J       => new String'("public"),
     Raise_A        => new String'("raise"),
     Range_A        => new String'("range"),
     Record_A       => new String'("record"),
     Rem_A          => new String'("rem"),
     Renames_A      => new String'("renames"),
     Requeue_A      => new String'("requeue"),
     Return_AJ      => new String'("return"),
     Reverse_A      => new String'("reverse"),
     Select_A       => new String'("select"),
     Separate_A     => new String'("separate"),
     Short_J        => new String'("short"),
     Static_J       => new String'("static"),
     Subtype_A      => new String'("subtype"),
     Super_J        => new String'("super"),
     Switch_J       => new String'("switch"),
     Synchronized_J => new String'("synchronized"),
     Tagged_A       => new String'("tagged"),
     Task_A         => new String'("task"),
     Terminate_A    => new String'("terminate"),
     Then_A         => new String'("then"),
     This_J         => new String'("this"),
     Throw_J        => new String'("throw"),
     Throws_J       => new String'("throws"),
     Transient_J    => new String'("transient"),
     Try_J          => new String'("try"),
     Type_A         => new String'("type"),
     Until_A        => new String'("until"),
     Use_A          => new String'("use"),
     Void_J         => new String'("void"),
     Volatile_J     => new String'("volatile"),
     When_A         => new String'("when"),
     While_AJ       => new String'("while"),
     With_A         => new String'("with"),
     Xor_A          => new String'("xor"));

  Delimiter_Image: constant array (Delimiter) of String_Pointer :=
    (Colon_AJ                   => new String'(":"),
     Comma_AJ                   => new String'(","),
     Dot_AJ                     => new String'("."),
     Semicolon_AJ               => new String'(";"),
     Tick_A                     => new String'("'"),
     LeftBrace_J                => new String'("{"),
     RightBrace_J               => new String'("}"),
     LeftBracket_J              => new String'("["),
     RightBracket_J             => new String'("]"),
     Left_Parenthesis_AJ        => new String'("("),
     Right_Parenthesis_AJ       => new String'(")"),
     Concatenate_A              => new String'("&"),
     And_J                      => new String'("&"),
     Alternative_A              => new String'("|"),
     Or_J                       => new String'("|"),
     Assignment_J               => new String'("="),
     Conditional_J              => new String'("?"),
     Equal_A                    => new String'("="),
     Greater_Than_AJ            => new String'(">"),
     Less_Than_AJ               => new String'("<"),
     Complement_J               => new String'("~"),
     Not_J                      => new String'("!"),
     Xor_J                      => new String'("^"),
     Plus_AJ                    => new String'("+"),
     Minus_AJ                   => new String'("-"),
     Times_AJ                   => new String'("*"),
     Divide_AJ                  => new String'("/"),
     Remainder_J                => new String'("%"),
     Arrow_A                    => new String'("=>"),
     Assignment_A               => new String'(":="),
     Double_Dot_A               => new String'(".."),
     Exponentiate_A             => new String'("**"),
     Equal_J                    => new String'("=="),
     Not_Equal_A                => new String'("/="),
     NotEqual_J                 => new String'("!="),
     Greater_Equal_AJ           => new String'(">="),
     Less_Equal_AJ              => new String'("<="),
     Left_Label_Bracket_A       => new String'("<<"),
     Right_Label_Bracket_A      => new String'(">>"),
     Box_A                      => new String'("<>"),
     Increment_J                => new String'("++"),
     Decrement_J                => new String'("--"),
     LeftShift_J                => new String'("<<"),
     RightShift_J               => new String'(">>"),
     UnsignedRightShift_J       => new String'(">>>"),
     ShortCutOr_J               => new String'("||"),
     ShortCutAnd_J              => new String'("&&"),
     PlusAssign_J               => new String'("+="),
     MinusAssign_J              => new String'("-="),
     TimesAssign_J              => new String'("*="),
     DivideAssign_J             => new String'("/="),
     RemainderAssign_J          => new String'("%="),
     AndAssign_J                => new String'("&="),
     OrAssign_J                 => new String'("|="),
     XorAssign_J                => new String'("^="),
     LeftShiftAssign_J          => new String'("<<="),
     RightShiftAssign_J         => new String'(">>="),
     UnsignedRightShiftAssign_J => new String'(">>>="));

  Java_Literal_Image: constant array (Null_J .. True_J) of String_Pointer :=
    (Null_J  => Reserved_Word_Image (Null_A),
     False_J => new String'("false"),
     True_J  => new String'("true"));

  Documentation_Tag_Image: constant array (Proper_Tag) of String_Pointer :=
    (See_Tag       => new String'("see"),
     Author_Tag    => new String'("author"),
     Version_Tag   => new String'("version"),
     Param_Tag     => new String'("param"),
     Return_Tag    => Reserved_Word_Image (Return_AJ),
     Exception_Tag => Reserved_Word_Image (Exception_A));

  -----------------------------------------------------------------------
  -- Visible token operations

  function Name (Item: Token) return Token_Name is
  begin
    return Item.Core.Name;
  end Name;

  function Image (Item: Token) return String is
  begin
    case Item.Core.Name is
      when Reserved_Word =>
        return Reserved_Word_Image (Item.Core.Name).all;
      when Delimiter =>
        return Delimiter_Image (Item.Core.Name).all;
      when Null_J .. True_J =>
        return Java_Literal_Image (Item.Core.Name).all;
      when Comment_A =>
        return "--" & Item.Core.Image.all;
      when EndOfLineComment_J =>
        return "//" & Item.Core.Image.all;
      when TraditionalComment_J =>
        return "/*" & Item.Core.Image.all & "*/";
      when CommentHead_J =>
        return "/*" & Item.Core.Image.all;
      when CommentTail_J | DocumentationTail_J =>
        return Item.Core.Image.all & "*/";
      when Documentation_J =>
        return "/**" & Item.Core.Image.all & "*/";
      when DocumentationHead_J =>
        return "/**" & Item.Core.Image.all;
      when String_AJ =>
        return '"' & Item.Core.Image.all & '"';
      when Character_AJ =>
        return ''' & Item.Core.Image.all & ''';
      when others =>
        return Item.Core.Image.all;
    end case;
  end Image;

  function Is_Operator (Item: Token) return Boolean is
  begin
    case Item.Core.Name is
      when Colon_AJ =>
        return Program_Language = Java;
      when Concatenate_A | And_J |
           Or_J .. Remainder_J |
           Exponentiate_A ..Less_Equal_AJ |
           Increment_J .. UnsignedRightShiftAssign_J =>
        return True;
      when others =>
        return False;
    end case;
  end Is_Operator;

  function Base (Item: Token) return Number_Base is
  begin
    return Item.Core.Base;
  end Base;

  function Tag (Item: Token) return Documentation_Tag is
  begin
    return Item.Core.Tag;
  end Tag;

  function Tag_Pos (Item: Token) return Natural is
  begin
    return Item.Core.Pos;
  end Tag_Pos;

  function Image (Tag: Proper_Tag) return String is
  begin
    return '@' & Documentation_Tag_Image (Tag).all;
  end Image;

  function Error (Item: Token) return Token_Error is
  begin
    return Item.Core.Error;
  end Error;

  function First (Item: Token) return Positive is
  begin
    return Item.Core.First;
  end First;

  function Last (Item: Token) return Positive is
  begin
    return Item.Core.Last;
  end Last;

  function Line (Item: Token) return Positive is
  begin
    return Item.Core.Line;
  end Line;

  function Col (Item: Token) return Positive is
  begin
    return Item.Core.Col;
  end Col;

  -----------------------------------------------------------------------
  -- Mappings and decisions for parsing

  function To_Lower (Item: String) return String renames Characters.Handling.To_Lower;

  function Map (Word: String) return String is
    -- Is Word a reserved word?
  begin
    case Program_Language is
      when Unknown |
           Ada     => return To_Lower (Word);
      when Java    => return Word;  -- Java is case-sensitive
    end case;
  end Map;

  function Is_Alphanumeric      (Item: Character) return Boolean renames Characters.Handling.Is_Alphanumeric;
  function Is_Decimal_Digit     (Item: Character) return Boolean renames Characters.Handling.Is_Decimal_Digit;
  function Is_Hexadecimal_Digit (Item: Character) return Boolean renames Characters.Handling.Is_Hexadecimal_Digit;
  function Is_Control           (Item: Character) return Boolean renames Characters.Handling.Is_Control;
  function Is_Graphic           (Item: Character) return Boolean renames Characters.Handling.Is_Graphic;

  function Is_Alphanumeric_or_Dollar (Item: Character) return Boolean is
  begin
    return Is_Alphanumeric (Item) or (Program_Language = Java and Item ='$');
  end Is_Alphanumeric_or_Dollar;

  function Is_Decimal_Digit_or_Underline (Item: Character) return Boolean is
  begin
    return Is_Decimal_Digit (Item) or Item = '_';
  end Is_Decimal_Digit_or_Underline;

  function Is_Hexadecimal_Digit_or_Underline (Item: Character) return Boolean is
  begin
    return Is_Hexadecimal_Digit (Item) or Item = '_';
  end Is_Hexadecimal_Digit_or_Underline;

  -----------------------------------------------------------------------
  -- Input stream processing (a set of global variables is used rather
  -- than subprogram parameters)

  Input_Location: Positive;        -- the next position to read
  Input_String  : String_Pointer;  -- in this string

  Current_Line,            -- the next position in
  Current_Col : Positive;  -- the file
  Tab_Size    : Positive;

  type Java_Comment is (None, Simple, Document);
  Comment_to_Complete: Java_Comment;

  Exponent_Present: Boolean;

  Next_Char: Character;  -- the character just read and to be processed

  type Error_Occurrence (Occurred: Boolean := False) is record  -- for bad token
    case Occurred is
      when False => null;
      when True  => Name: Token_Error;
    end case;
  end record;

  function "or" (Left, Right: Error_Occurrence) return Error_Occurrence is
  begin
    if Left.Occurred then
      return Left;
    elsif Right.Occurred then
      return Right;
    else
      return (Occurred => False);
    end if;
  end "or";

  Last_Error: Error_Occurrence;

  -- EOL (end of line) is defined as {CR}LF.
  CR : Character renames Characters.Latin_1.CR;   -- carriage return
  LF : Character renames Characters.Latin_1.LF;   -- line feed
  EOT: Character renames Characters.Latin_1.EOT;  -- end of input string
  HT : Character renames Characters.Latin_1.HT;   -- horizontal tabulation

  -----------------------------------------------------------------------
  -- A simple stack remembering the last two column numbers (needed to
  -- go back across LF and HT) - (we go back at most two characters)

  package Stack is

    procedure Push (Col: in  Positive);
    procedure Pop  (Col: out Positive);

  end Stack;

  package body Stack is

    Store: array (Boolean) of Positive;
    Top  : Boolean := Store'First;

    procedure Push (Col: in Positive) is
    begin
      Top := not Top;
      Store (Top) := Col;
    end Push;

    procedure Pop (Col: out Positive) is
    begin
      Col := Store (Top);
      Top := not Top;
    end Pop;

  end Stack;

  -----------------------------------------------------------------------
  -- Reading input stream

  function Get_Next_Char return Character is
    -- Return next character from input stream and update position.
    Char: Character;
  begin
    Stack.Push (Current_Col);
    if Input_Location > Input_String'Last then
      Input_Location := Input_Location + 1;  -- simplifies counting
      return EOT;
    end if;
    if Input_String (Input_Location) = LF then
      Current_Line := Current_Line + 1;
      Current_Col  := 1;
    elsif Input_String (Input_Location) = HT then
      Current_Col := Current_Col + Tab_Size - (Current_Col - 1) mod Tab_Size;
    else
      Current_Col := Current_Col + 1;
    end if;
    Char := Input_String (Input_Location);
    Input_Location := Input_Location + 1;
    return Char;
  end Get_Next_Char;

  procedure Put_Back_Last_Character is
  begin
    Input_Location := Input_Location - 1;
    if Input_Location <= Input_String'Last and then Input_String (Input_Location) = LF then
      Current_Line := Current_Line - 1;
    end if;
    Stack.Pop (Current_Col);
  end Put_Back_Last_Character;

  procedure Skip_White_Space is
  begin
    loop
      Next_Char := Get_Next_Char;
      exit when Next_Char = EOT or else
               (Next_Char /= ' ' and then not Is_Control (Next_Char));
    end loop;
  end Skip_White_Space;

  function Look_Ahead_is (Expected: Character) return Boolean is
    -- If the next character is the expected one, it is kept;
    -- else it stays in the input stream.
    Save_Char: constant Character := Next_Char;
  begin
    Next_Char := Get_Next_Char;
    if Next_Char = Expected then
      return True;
    else
      Put_Back_Last_Character;
      Next_Char := Save_Char;
      return False;
    end if;
  end Look_Ahead_is;

  -----------------------------------------------------------------------
  -- Basic constructors

  function Construct (Name: Delimiter) return Token is
    Whom: Token_Core (Name);
  begin
    Whom.First := Input_Location - Delimiter_Image (Name).all'Length;
    Whom.Last  := Input_Location - 1;
    Whom.Col   := Current_Col    - Delimiter_Image (Name).all'Length;
    Whom.Line  := Current_Line;
    return (Finalization.Controlled with Whom);
  end Construct;

  function Construct_Bad_Token (For_Error: Error_Occurrence;
                                Start_Location, End_Location,
                                Start_Line, Start_Col: Positive) return Token is
  begin
    return (Finalization.Controlled with
            Core => (Name  => Bad_Token_AJ,
                     Image => new String'(Input_String (Start_Location .. End_Location)),
                     Error => For_Error.Name,
                     First => Start_Location,
                     Last  => End_Location,
                     Line  => Start_Line,
                     Col   => Start_Col));
  end Construct_Bad_Token;

  function Construct (Name : Token_Name;
                      Image: String;
                      First, Last,
                      Line, Col: Positive;
                      Base: Natural := 0;
                      Pos : Natural := 0;
                      Tag : Documentation_Tag := No_Tag) return Token is
    -- Must only be called for tokens with an image.
    -- Base must be in Number_Base for based numbers.
    -- Pos and Tag are only evaluated for Java documentation comments.
    Result: Token_Core (Name);
  begin
    Result.Image  := new String'(Image);
    Result.First  := First;
    Result.Last   := Last;
    Result.Line   := Line;
    Result.Col    := Col;
    if Base in Number_Base then
      Result.Base := Base;
    elsif Name in Documentation then
      Result.Tag  := Tag;
      Result.Pos  := Pos;
    end if;
    return (Finalization.Controlled with Result);
  end Construct;

  -----------------------------------------------------------------------
  -- Constructing comments

  function Get_Line return String is
    -- Read all characters up to and including the next EOL,
    -- and return them without the EOL.
    -- Attention: Illegal control characters are not recognized.
    Start: constant Positive := Input_Location;
    Last: Positive;
  begin
    loop
      Next_Char := Get_Next_Char;
      exit when Next_Char = LF or Next_Char = EOT;
    end loop;
    Last := Input_Location - 2;
    if Input_String (Last) = CR then
      Last := Last - 1;
    end if;
    return Input_String (Start .. Last);
  end Get_Line;

  function Construct_Comment_Line return Token is
    -- Return the rest of the characters on the current line as a
    -- comment token.
    -- Note that the introducing string "--" respectively "//" is not
    -- stored in the token (it will be added by the Image function),
    -- but included in the position information.
    Start_Location: constant Positive := Input_Location - 2;
    Start_Line    : constant Positive := Current_Line;
    Start_Col     : constant Positive := Current_Col - 2;
    Rest_of_Line  : constant String   := Get_Line;
    Comment: Token_Name;
  begin
    if Input_String (Start_Location .. Start_Location + 1) = "--" then
      Comment := Comment_A;
    else
      Comment := EndOfLineComment_J;
    end if;
    return Construct (Name  => Comment,
                      Image => Rest_of_Line,
                      First => Start_Location,
                      Last  => Start_Location + Rest_of_Line'Length + 1,
                      Line  => Start_Line,
                      Col   => Start_Col);
  end Construct_Comment_Line;

  function Get_Rest_of_Comment return String is
    -- Get all characters until EOL or */ whichever is first.
    Start_Location: constant Positive := Input_Location;
    End_Location  :          Positive;
  begin
    loop
      Next_Char := Get_Next_Char;
      exit when  Next_Char = LF or Next_Char = EOT or
                (Next_Char = '*' and then Look_Ahead_is ('/'));
    end loop;
    if Input_Location > Input_String'Last then
      End_Location := Input_String'Last;
    else
      End_Location := Input_Location - 1;
    end if;
    if Input_String (End_Location) = LF then
      End_Location := End_Location - 1;
      if Input_String (End_Location) = CR then
        End_Location := End_Location - 1;
      end if;
    end if;
    return Input_String (Start_Location .. End_Location);
  end Get_Rest_of_Comment;

  procedure Find_Tag (Token: in Comment; Text: in String;
                      Tag: out Documentation_Tag; Found_At: out Natural) is
    Tag_Length: Positive;
  begin
    if Token in Documentation then
      Found_At := Strings.Fixed.Index (Source => Text, Pattern => "@");
      if Found_At > 0 then
        for This_Tag in Proper_Tag loop
          Tag_Length := Documentation_Tag_Image (This_Tag)'Length;
          if Text'Last >= Found_At + Tag_Length and then
             Text (Found_At + 1 .. Found_At + Tag_Length) = Documentation_Tag_Image (This_Tag).all and then
             (Text'Last = Found_At + Tag_Length or else
              not Is_Alphanumeric (Text (Found_At + Tag_Length + 1))) then
            Tag := This_Tag;
            return;
          end if;
        end loop;
      end if;
    end if;
    Tag      := No_Tag;
    Found_At := 0;
  end Find_Tag;

  function Construct_other_Java_Comment return Token is
    -- Any comment starting with "/*".
    Start_Location : constant Positive := Input_Location - 2;
    Start_Line     : constant Positive := Current_Line;
    Start_Col      : constant Positive := Current_Col - 2;
    Rest_of_Comment: constant String   := Get_Rest_of_Comment;
    Comment    : Token_Name;
    First, Last: Positive;
    Pos: Natural;
    Tag: Documentation_Tag;
  begin
    if Rest_of_Comment'Length >= 2 and then  -- Attention: /*/ is not a full comment.
       Rest_of_Comment (Rest_of_Comment'Last - 1 .. Rest_of_Comment'Last) = "*/" then  -- a whole comment
      if Rest_of_Comment'Length > 2 and then Rest_of_Comment (Rest_of_Comment'First) = '*' then
        Comment := Documentation_J;
        First   := Rest_of_Comment'First + 1;
      else
        Comment := TraditionalComment_J;
        First   := Rest_of_Comment'First;
      end if;
      Last := Rest_of_Comment'Last  - 2;
    else  -- begin of a comment
      if Input_String'Last >= Start_Location + 2 and then
         Input_String (Start_Location + 2) = '*' then
        Comment_to_Complete := Document;
        Comment := DocumentationHead_J;
        First   := Rest_of_Comment'First + 1;
      else
        Comment_to_Complete := Simple;
        Comment := CommentHead_J;
        First   := Rest_of_Comment'First;
      end if;
      Last := Rest_of_Comment'Last;
    end if;
    Find_Tag (Comment, Rest_of_Comment (First .. Last), Tag, Pos);
    return Construct (Name  => Comment,
                      Image => Rest_of_Comment (First .. Last),
                      First => Start_Location,
                      Last  => Start_Location + Rest_of_Comment'Length + 1,
                      Line  => Start_Line,
                      Col   => Start_Col,
                      Tag   => Tag,
                      Pos   => Pos);
  end Construct_other_Java_Comment;

  function Complete_Java_Comment return Token is
    Start_Location : constant Positive := Input_Location;
    Start_Line     : constant Positive := Current_Line;
    Start_Col      : constant Positive := Current_Col;
    Rest_of_Comment: constant String   := Get_Rest_of_Comment;
    Comment    : Token_Name;
    First, Last: Positive;
    Pos: Natural;
    Tag: Documentation_Tag;
  begin
    if Rest_of_Comment'Length >= 2 and then
       Rest_of_Comment (Rest_of_Comment'Last - 1 .. Rest_of_Comment'Last) = "*/" then  -- end of a comment
      case Comment_to_Complete is
        when None     => null;   -- impossible
        when Simple   => Comment := CommentTail_J;        Comment_to_Complete := None;
        when Document => Comment := DocumentationTail_J;  Comment_to_Complete := None;
      end case;
      First := Rest_of_Comment'First;
      Last  := Rest_of_Comment'Last - 2;
    else  -- middle part of a comment
      case Comment_to_Complete is
        when None     => null;   -- impossible
        when Simple   => Comment := CommentBody_J;
        when Document => Comment := DocumentationBody_J;
      end case;
      First := Rest_of_Comment'First;
      Last  := Rest_of_Comment'Last;
    end if;
    Find_Tag (Comment, Rest_of_Comment (First .. Last), Tag, Pos);
    return Construct (Name  => Comment,
                      Image => Rest_of_Comment (First .. Last),
                      First => Start_Location,
                      Last  => Start_Location + Rest_of_Comment'Length + 1,
                      Line  => Start_Line,
                      Col   => Start_Col,
                      Tag   => Tag,
                      Pos   => Pos);
  end Complete_Java_Comment;

  -----------------------------------------------------------------------

  generic
    with function is_Expected (Item: Character) return Boolean;
    with function Underline_is_not_Allowed      return Boolean;
  procedure Analyse_Sequence;

  procedure Analyse_Sequence is
    Start_Location: constant Positive := Input_Location - 1;  -- first character already read
    Last_was_Underline: Boolean := False;
    This_is_Underline : Boolean := False;
  begin
    if (Underline_is_not_Allowed or Program_Language = Ada) and Input_String (Start_Location) = '_' then
      Last_Error := (Occurred => True,
                     Name     => Illegal_Underline);
    else
      Last_Error := (Occurred => False);
    end if;
    loop
      Next_Char := Get_Next_Char;
      This_is_Underline := Next_Char = '_';
      if (Underline_is_not_Allowed and This_is_Underline) or 
         (Program_Language = Ada and Last_was_Underline and This_is_Underline) then
        Last_Error := (Occurred => True,
                       Name     => Illegal_Underline);
      end if;
      exit when not This_is_Underline and
                not is_Expected (Next_Char);
      Last_was_Underline := This_is_Underline;
    end loop;
    if Program_Language = Ada and Last_was_Underline then
      Last_Error := (Occurred => True,
                     Name     => Illegal_Underline);
    end if;
    Put_Back_Last_Character;
  end Analyse_Sequence;

  function Underline_is_not_Allowed_for_Java return Boolean is
  begin
    return Program_Language = Java;
  end Underline_is_not_Allowed_for_Java;

  procedure Analyse_Identifier    is new Analyse_Sequence (Is_Alphanumeric_or_Dollar, Underline_is_not_Allowed => False);
  procedure Analyse_Numeral       is new Analyse_Sequence (Is_Decimal_Digit         , Underline_is_not_Allowed_for_Java);
  procedure Analyse_Based_Numeral is new Analyse_Sequence (Is_Hexadecimal_Digit     , Underline_is_not_Allowed_for_Java);

  procedure Analyse_Based_Numeral (Base: in Number_Base) is
    Start_Location: constant Positive := Input_Location - 1;  -- first character already read
  begin
    Analyse_Based_Numeral;
    for Pos in Start_Location .. Input_Location - 1 loop
      if Input_String (Pos) /= '_' and then
         Number_Base'Value ("16#" & Input_String (Pos) & '#') >= Base then
        Last_Error := (Occurred => True,
                       Name     => Illegal_Extended_Digit);
        return;
      end if;
    end loop;
  end Analyse_Based_Numeral;

  procedure Analyse_Exponent (Constructing_an_Integer: in Boolean := False) is
    -- Exponent string may be empty.
    -- A negative sign for an integer is accepted, but marked as illegal.
    Illegal_Sign: Boolean := False;
  begin
    Last_Error := (Occurred => False);
    if Look_Ahead_is ('E') or else Look_Ahead_is ('e') then
      Exponent_Present := True;
      if Look_Ahead_is ('+') or else Look_Ahead_is ('-') then
        Illegal_Sign := Constructing_an_Integer and Next_Char = '-';
      end if;
      if Is_Decimal_Digit_or_Underline (Get_Next_Char) then
        Analyse_Numeral;
        if Illegal_Sign then
          Last_Error := (Occurred => True,
                         Name     => Illegal_Literal);
        end if;
      else
        Last_Error := (Occurred => True,
                       Name     => Illegal_Literal);
        Put_Back_Last_Character;
      end if;
    else
      Exponent_Present := False;
    end if;
  end Analyse_Exponent;

  function Construct_Identifier_or_Keyword return Token is
    Start_Location: constant Positive := Input_Location - 1;
    Start_Line    : constant Positive := Current_Line;
    Start_Col     : constant Positive := Current_Col - 1;
    First, Last   : Reserved_Word;
    function "=" (Left: Token_Name; Right: String) return Boolean is
    begin
      if Left in Reserved_Word then
        return Reserved_Word_Image (Left).all = Map (Right);
      else
        return Java_Literal_Image (Left).all = Right;
      end if;
    end "=";
    function Construct_Keyword (Word: in Token_Name) return Token is
      Result: Token_Core (Word);
    begin
      Result.First := Start_Location;
      Result.Last  := Input_Location - 1;
      Result.Line  := Start_Line;
      Result.Col   := Start_Col;
      return (Finalization.Controlled with Result);
    end Construct_Keyword;
  begin
    Analyse_Identifier;
    if Last_Error.Occurred then
      return Construct_Bad_Token (Last_Error,
                                  Start_Location, Input_Location - 1,
                                  Start_Line, Start_Col);
    end if;
    -- Check for reserved words:
    case Input_String (Start_Location) is
      when 'A'|'a' => First := Abort_A;    Last := At_A;
      when 'B'|'b' => First := Begin_A;    Last := Byte_J;
      when 'C'|'c' => First := Case_AJ;    Last := Continue_J;
      when 'D'|'d' => First := Declare_A;  Last := Double_J;
      when 'E'|'e' => First := Else_AJ;    Last := Extends_J;
      when 'F'|'f' => First := Final_J;    Last := Function_A;
      when 'G'|'g' => First := Generic_A;  Last := Goto_AJ;
      when 'I'|'i' => First := If_AJ;      Last := Is_A;
      when 'L'|'l' => First := Limited_A;  Last := Loop_A;
      when 'M'|'m' => First := Mod_A;      Last := Mod_A;
      when 'N'|'n' => First := Native_J;   Last := Null_A;
      when 'O'|'o' => First := Of_A;       Last := Out_A;
      when 'P'|'p' => First := Package_AJ; Last := Public_J;
      when 'R'|'r' => First := Raise_A;    Last := Reverse_A;
      when 'S'|'s' => First := Select_A;   Last := Synchronized_J;
      when 'T'|'t' => First := Tagged_A;   Last := Type_A;
      when 'U'|'u' => First := Until_A;    Last := Use_A;
      when 'V'|'v' => First := Void_J;     Last := Volatile_J;
      when 'W'|'w' => First := When_A;     Last := With_A;
      when 'X'|'x' => First := Xor_A;      Last := Xor_A;
      when others  => First := Xor_A;      Last := Abort_A;  -- null range
    end case;
    for Word in First .. Last loop
      if Word = Input_String (Start_Location .. Input_Location - 1) then
        if Is_Reserved (Word, Program_Language) then
          return Construct_Keyword (Word);
        else
          exit;
        end if;
      end if;
    end loop;
    if Program_Language = Java then
      for Word in Null_J .. True_J loop
        if Word = Input_String (Start_Location .. Input_Location - 1) then
          return Construct_Keyword (Word);
        end if;
      end loop;
    end if;
    -- Not a keyword.
    return Construct (Name  => Identifier_AJ,
                      Image => Input_String (Start_Location .. Input_Location - 1),
                      First => Start_Location,
                      Last  => Input_Location - 1,
                      Line  => Start_Line,
                      Col   => Start_Col);
  end Construct_Identifier_or_Keyword;

  function Construct_Based_Number (Start_Location, Start_Line, Start_Col: in Positive) return Token is
    -- Base is already read, we have just found the number sign (Ada) or leading
    -- 0[x] (Java).
    -- The next character read must be the first of the numeral.
    Base: Number_Base;
    Base_Error, Error_in_Whole_Part, Error_in_Decimal_Part: Error_Occurrence;
    Decimal_Dot_Present: Boolean := False;
    procedure Check_Base is
    begin
      if Program_Language = Java then
        if Input_String (Start_Location + 1) = 'x' or Input_String (Start_Location + 1) = 'X' then
          Base := 16;
        else
          Base :=  8;
        end if;
      else
        Base := Integer'Value (Input_String (Start_Location .. Input_Location - 2));
      end if;
      Base_Error := (Occurred => False);
    exception  -- only for Ada
      when others =>
        Base := 16;  -- do not check digits
        Base_Error := (Occurred => True,
                       Name     => Illegal_Base);
    end Check_Base;
    function Construct_Bad_Token (For_Error: Error_Occurrence) return Token is
    begin
      return Construct_Bad_Token (For_Error,
                                  Start_Location, Input_Location - 1,
                                  Start_Line, Start_Col);
    end Construct_Bad_Token;
    function Construct (Name: Token_Name) return Token is
    begin
      return Construct (Name  => Name,
                        Image => Input_String (Start_Location .. Input_Location - 1),
                        Base  => Base,
                        First => Start_Location,
                        Last  => Input_Location - 1,
                        Line  => Start_Line,
                        Col   => Start_Col);
    end Construct;
  begin
    Check_Base;
    -- We show with an example string (Ada) how far we have analysed the input
    -- and which characters are expected next.
    -- 16#
    -- Expected:  extended digits
    -- Otherwise: error
    if Is_Hexadecimal_Digit_or_Underline (Get_Next_Char) then
      -- 16#F
      -- Expected:  further extended digits
      --            dot
      --            double dot
      --            number sign
      -- Otherwise: error
      Analyse_Based_Numeral (Base);  -- whole part
      Error_in_Whole_Part := Last_Error;
      -- We are now at the end of the whole part.
    else  -- an unexpected character c
      -- 16#c
      -- We treat this case as though we had read an erroneous whole part.
      Error_in_Whole_Part := (Occurred => True,
                              Name     => Illegal_Literal);
      Put_Back_Last_Character;  -- reset to position of unexpected character
    end if;
    -- Ada:  16#F where F stands for the (correct or erroneous) whole part.
    --       Expected:  single dot (it's a real)
    --                  number sign (it's an integer)
    --       Otherwise: error
    -- Java: 0[x]F with F as above
    --       Expected:  length character (it's a long integer)
    --                  nothing (it's an integer)
    case Program_Language is
      when Ada =>
        if Look_Ahead_is ('.') then
          Decimal_Dot_Present := True;
          if Look_Ahead_is ('.') then  -- a double dot, so it's an error
            -- 16#F..
            Input_Location := Input_Location - 2;  -- reset to position
            Current_Col    := Current_Col    - 2;  -- of first dot
            return Construct_Bad_Token ((Occurred => True,
                                         Name     => Missing_Base_Quote));
          elsif Look_Ahead_is ('#') then  -- missing decimal part
            -- 16#F.#
            -- Check for exponent.
            Error_in_Decimal_Part := (Occurred => True,
                                      Name     => Illegal_Literal);
          elsif Is_Hexadecimal_Digit_or_Underline (Get_Next_Char) then
            -- 16#F.F
            -- Expected:  number sign
            --            further extended digits
            -- Otherwise: error
            Analyse_Based_Numeral (Base);  -- decimal part
            Error_in_Decimal_Part := Last_Error;
            if not Look_Ahead_is ('#') then
              -- 16#F.Fc
              -- Error
              return Construct_Bad_Token ((Occurred => True,
                                           Name     => Missing_Base_Quote));
            end if;
            -- 16#F.F#
            -- Check for exponent.
          else
            -- 16#F.c
            -- Error
            Put_Back_Last_Character;  -- reset to position after dot
            return Construct_Bad_Token ((Occurred => True,
                                         Name     => Missing_Base_Quote));
          end if;
          -- 16#F.F#
          -- Check for exponent.
        elsif Look_Ahead_is ('#') then
          -- 16#F#
          -- Check for exponent.
          null;
        else
          -- 16#Fc
          -- Error
          return Construct_Bad_Token ((Occurred => True,
                                       Name     => Missing_Base_Quote));
        end if;
        -- 16#F.F#
        -- Based number, integer or real, possibly with errors, but
        -- without exponent.
        Analyse_Exponent (Constructing_an_Integer => not Decimal_Dot_Present);
        if Last_Error.Occurred or Base_Error.Occurred or
           Error_in_Whole_Part.Occurred or Error_in_Decimal_Part.Occurred then
          return Construct_Bad_Token (Last_Error or Base_Error or
                                      Error_in_Whole_Part or Error_in_Decimal_Part);
        elsif Decimal_Dot_Present then
          return Construct (Name  => Based_Real_A);
        else
          return Construct (Name  => Based_Integer_AJ);
        end if;
      when Java =>
        if Look_Ahead_Is ('l') or else Look_Ahead_Is ('L') then
          if not Error_In_Whole_Part.Occurred then
            return Construct (Name  => BasedLongInteger_J);
          end if;
        elsif not Error_In_Whole_Part.Occurred then
          return Construct (Name  => Based_Integer_AJ);
        end if;
        return Construct_Bad_Token (Error_In_Whole_Part);
      when Unknown =>
        null;  -- deliberately let it crash
    end case;
  end Construct_Based_Number;

  function Construct_Number return Token is
    -- The first character has already been read. It is a digit or a dot
    -- (for Java lazy reals only).
    -- When there are errors in numbers, it's rather arbitrary when to stop
    -- parsing the input.
    -- Take e.g. "12.3.56.0". What might have been meant is the range
    -- "12.3..56.0".
    -- Or what might be the intent of "12.0_e-4", "16#123#E4#",
    -- "12#F23e4"?
    -- All characters are taken until a number is complete or until this
    -- is definitely an error.
    -- We do not strive to giving an extensive error analysis and hints
    -- to Ada  or Java newbies.
    -- Note: We cannot use the 'Value attribute to check whether a legal
    --       integer literal has been parsed because constants may be
    --       outside the number range any type can represent.
    -- Problem in Java:
    --   For a digit sequence starting with 0, we can only determine at the
    --   end whether it is an octal or a real number, depending on whether
    --   there is an exponent or not:
    --   07:    legal octal,
    --   08:    illegal octal;
    --   08E+2: legal real.
    -- Algorithm:
    -- ----------
    -- A number can consist of several parts: numerals; dots; base, exponent,
    -- length (Java only) indicators. We analyse them sequentially, after
    -- each step deciding what may follow.
    Start_Location: constant Positive := Input_Location - 1;
    Start_Line    : constant Positive := Current_Line;
    Start_Col     : constant Positive := Current_Col - 1;
    Error_in_First_Numeral, Error_in_Fractional_Part: Error_Occurrence;
    Java_D_Specified, Java_F_Specified: Boolean := False;
    function Construct_Bad_Token (For_Error: Error_Occurrence) return Token is
    begin
      return Construct_Bad_Token (For_Error,
                                  Start_Location, Input_Location - 1,
                                  Start_Line, Start_Col);
    end Construct_Bad_Token;
    function Construct (Name: Token_Name) return Token is
    begin
      return Construct (Name  => Name,
                        Image => Input_String (Start_Location .. Input_Location - 1),
                        First => Start_Location,
                        Last  => Input_Location - 1,
                        Line  => Start_Line,
                        Col   => Start_Col);
    end Construct;
  begin
    -- We immediately separate Java hexadecimal numbers.
    if (Program_Language = Java and Next_Char = '0') and then
       (Look_Ahead_is ('x') or else Look_Ahead_is ('X')) then
      return Construct_Based_Number (Start_Location, Start_Line, Start_Col);
    end if;
    -- It's any Ada number or a Java decimal or octal number.
    Analyse_Numeral;
    -- The first numeral might be:
    -- Ada:  A decimal integer literal, the whole part of a decimal real
    --       literal, or the base of a based literal (integer or real).
    -- Java: A decimal or octal integer literal, the whole part of a decimal
    --       real literal (where the fractional part may be missing if there
    --       is an exponent).
    -- Since we can discriminate between Java octal and real only after we
    -- checked for an exponent, we postpone the decision until it is due.
    Error_in_First_Numeral := Last_Error;
    if Look_Ahead_is ('.') then
      if Look_Ahead_is ('.') then  -- a double dot, so it's an integer
        Input_Location := Input_Location - 2;  -- reset to position
        Current_Col    := Current_Col    - 2;  -- of first dot
        return Construct (Name  => Integer_AJ);
      else  -- a single dot, so it's a real
        if Is_Decimal_Digit_or_Underline (Get_Next_Char) then
          Analyse_Numeral;  -- fractional part
          Error_in_Fractional_Part := Last_Error;
        else
          Put_Back_Last_Character;  -- reset to position after dot
          Error_in_Fractional_Part := (Occurred => True,
                                       Name     => Illegal_Literal);
        end if;
        Analyse_Exponent;
        if Program_Language = Java then
          Java_D_Specified :=  Look_Ahead_is ('d') or else Look_Ahead_is ('D');
          Java_F_Specified := not Java_D_Specified and then
                              (Look_Ahead_is ('f') or else Look_Ahead_is ('F'));
        end if;
        if Last_Error.Occurred or
           Error_in_First_Numeral.Occurred or Error_in_Fractional_Part.Occurred then
          return Construct_Bad_Token (Last_Error or
                                      Error_in_First_Numeral or Error_in_Fractional_Part);
        elsif Java_D_Specified then
          return Construct (Name  => DoubleNumber_J);
        elsif Java_F_Specified then
          return Construct (Name  => FloatNumber_J);
        else
          return Construct (Name  => Real_AJ);
        end if;
      end if;
    elsif Program_Language = Ada and then Look_Ahead_is ('#') then
      return Construct_Based_Number (Start_Location, Start_Line, Start_Col);
    else  -- it's an integer or Java real, float or double
      Analyse_Exponent (Constructing_an_Integer => Program_Language = Ada);
      if Program_Language = Java then
        Java_D_Specified :=  Look_Ahead_is ('d') or else Look_Ahead_is ('D');
        Java_F_Specified := not Java_D_Specified and then
                            (Look_Ahead_is ('f') or else Look_Ahead_is ('F'));
      end if;
      if Last_Error.Occurred then
        return Construct_Bad_Token (Last_Error);
      elsif Error_in_First_Numeral.Occurred then
        return Construct_Bad_Token (Error_in_First_Numeral);
      elsif Java_D_Specified then
        return Construct (Name  => DoubleNumber_J);
      elsif Java_F_Specified then
        return Construct (Name  => FloatNumber_J);
      elsif Exponent_Present and Program_Language = Java then
        return Construct (Name  => Real_AJ);
      elsif (not Exponent_Present and Program_Language = Java) then
        if Input_String (Start_Location) = '0' and Input_Location - Start_Location > 1 then
          -- Redo the whole analysis for an octal number, but first
          -- reset the input to the first character of the numeral.
          Input_Location := Start_Location + 1;
          Current_Col    := Start_Col      + 1;
          return Construct_Based_Number (Start_Location, Start_Line, Start_Col);
        elsif Look_Ahead_is ('l') or else Look_Ahead_is ('L') then
          return Construct (Name  => LongInteger_J);
        else
          return Construct (Name  => Integer_AJ);
        end if;
      else
        return Construct (Name  => Integer_AJ);
      end if;
    end if;
  end Construct_Number;

  -----------------------------------------------------------------------
  -- Constructing character and string literals

  function Construct_Character_Escape_Sequence return Token is
    -- JRM 3.3, 3.10.6.
    -- Note that the initial sequence '\ has already been processed.
    Save_Line : constant Positive  := Current_Line;
    Save_Col  : constant Positive  := Current_Col;
    Start_Col : constant Positive  := Current_Col - 2;
    Designator: constant Character := Get_Next_Char;
  begin
    if Look_Ahead_is (''') then  -- '\c' for any character c
      if Designator = 'b' or Designator = 't' or Designator = 'n' or Designator = 'f' or
         Designator = 'r' or Designator = '"' or Designator = ''' or Designator = '\' or
         Designator in '0' .. '7' then
        return Construct (Name  => Character_AJ,
                          Image => Input_String (Start_Col + 1 .. Input_Location - 2),
                          First => Start_Col,
                          Last  => Input_Location - 1,
                          Line  => Current_Line,
                          Col   => Start_Col);
      elsif Designator in '8' .. '9' then  -- illegal octal escape
        return Construct_Bad_Token ((Occurred => True,
                                     Name     => Illegal_Extended_Digit),
                                    Start_Col, Input_Location - 1,
                                    Current_Line, Start_Col);
      elsif Is_Graphic (Designator) then  -- illegal '\c' for any other c
        return Construct_Bad_Token ((Occurred => True,
                                     Name     => Non_Language_Character),
                                    Start_Col, Input_Location - 1,
                                    Current_Line, Start_Col);
      end if;
      -- Non-graphic enclosed.
      Put_Back_Last_Character;  -- put it back into the input stream
      Put_Back_Last_Character;  -- and also the '
    elsif Designator in '0' .. '9' then  -- '\d (d for decimal digit)
      -- There are at most three octal digits d and a final ' in an octal escape.
      -- We accept also decimal digits here and check for them later.
      -- Read the next three characters, but stop when a non-decimal is found.
      for C in 1 .. 3 loop
        Next_Char := Get_Next_Char;
        exit when not Is_Decimal_Digit (Next_Char);
      end loop;
      -- Now we have '\{d}c with any character c.
      if Next_Char = ''' then  -- '\{d}'
        -- Construct an Ada octal number and check its range.
        begin
          if Natural'Value ("8#" & Input_String (Start_Col + 2 .. Input_Location - 2) & '#') > 8#377# then
            raise Constraint_Error;
          end if;
          return Construct (Name  => Character_AJ,
                            Image => Input_String (Start_Col + 1 .. Input_Location - 2),
                            First => Start_Col,
                            Last  => Input_Location - 1,
                            Line  => Current_Line,
                            Col   => Start_Col);
        exception
          when Constraint_Error =>  -- out of range or non-octal digit
            return Construct_Bad_Token ((Occurred => True,
                                         Name     => Illegal_Extended_Digit),
                                        Start_Col, Input_Location - 1,
                                        Current_Line, Start_Col);
        end;
      end if;
      -- The final ' is missing.
      Put_Back_Last_Character;  -- decimal or not
      return Construct_Bad_Token ((Occurred => True,
                                   Name     => Illegal_Literal),
                                  Start_Col, Input_Location - 1,
                                  Current_Line, Start_Col);
--  elsif Designator = 'u' then  -- unicode escape
--    ...
    else
      -- Put Designator back into the input stream.
      Input_Location := Input_Location - 1;
    end if;
    -- Only '\ processed.
    -- A LF can have been enclosed, so reset the line and column number.
    Current_Line   := Save_Line;
    Current_Col    := Save_Col;
    if Look_Ahead_is (''') then  -- return '\'
      null;  -- consume the closing tick
    -- else return '\ only
    end if;
    return Construct_Bad_Token ((Occurred => True,
                                 Name     => Non_Language_Character),
                                Start_Col, Input_Location - 1,
                                Current_Line, Start_Col);
  end Construct_Character_Escape_Sequence;
 
  function Construct_Character_Literal_or_Tick return Token is
    -- A character literal is returned only when a legal character is
    -- enclosed between ticks.
    -- Note that the enclosing quotes ' are not stored in the token image
    -- (they will be added by the Image function), but included in the
    -- position information.
    -- When there is no pair of quotes or not a graphic character between
    -- them, a tick token is returned for Ada, a bad token for Java.
    Save_Line : constant Positive  := Current_Line;
    Save_Col  : constant Positive  := Current_Col;
    After_Tick: constant Character := Get_Next_Char;
  begin
    if Program_Language = Java and After_Tick = '\' then
      return Construct_Character_Escape_Sequence;
    elsif Look_Ahead_is (''') then
      if Is_Graphic (After_Tick) then
        if Program_Language = Java and (After_Tick = ''' or After_Tick = '"') then
          return Construct_Bad_Token ((Occurred => True,
                                       Name     => Non_Language_Character),
                                      Input_Location - 3, Input_Location - 1,
                                      Current_Line, Current_Col - 3);
        else
          return Construct (Name  => Character_AJ,
                            Image => Input_String (Input_Location - 2 .. Input_Location - 2),
                            First => Input_Location - 3,
                            Last  => Input_Location - 1,
                            Line  => Current_Line,
                            Col   => Current_Col - 3);
        end if;
      end if;
      -- Non-graphic enclosed.
      Put_Back_Last_Character;  -- second tick
    end if;
    -- Put the character after the first tick back into the input stream.
    Put_Back_Last_Character;
    if Program_Language = Ada then
      return Construct (Tick_A);
    else
      return Construct_Bad_Token ((Occurred => True,
                                   Name     => Non_Language_Character),
                                  Input_Location - 1, Input_Location - 1,
                                  Current_Line, Current_Col - 1);
    end if;
  end Construct_Character_Literal_or_Tick;

  function Construct_String_Literal return Token is
    -- Return all characters on the current line up to but not including
    -- a single string quote as a string token.
    -- Note that the enclosing string quotes '"' are not stored in the
    -- token (they will be added by the Image function), but included in
    -- the position information.
    -- When EOL is reached before the closing quote, a bad token is
    -- returned, which includes the initial string quote.
    Start_Location: constant Positive := Input_Location - 1;
    Start_Line    : constant Positive := Current_Line;
    Start_Col     : constant Positive := Current_Col - 1;
    Last: Positive;
  begin
    Last_Error := (Occurred => False);
    loop
      Next_Char := Get_Next_Char;
      if Next_Char = '"' then
        if not Look_Ahead_is ('"') then
          if Last_Error.Occurred then
            return Construct_Bad_Token (Last_Error,
                                        Start_Location, Input_Location - 1,
                                        Start_Line, Start_Col);
          else
            return Construct (Name  => String_AJ,
                              Image => Input_String (Start_Location + 1 .. Input_Location - 2),
                              First => Start_Location,
                              Last  => Input_Location - 1,
                              Line  => Start_Line,
                              Col   => Start_Col);
          end if;
        end if;
      elsif Next_Char = LF then
        Last := Input_Location - 2;
        if Input_String (Last) = CR then
          Last := Last - 1;
        end if;
        return Construct_Bad_Token ((Occurred => True,
                                     Name     => Missing_String_Quote),
                                    Start_Location, Last,
                                    Start_Line, Start_Col);
      elsif End_of_Input then
        return Construct_Bad_Token ((Occurred => True,
                                     Name     => Missing_String_Quote),
                                    Start_Location, Input_Location - 2,
                                    Start_Line, Start_Col);
      elsif not Is_Graphic (Next_Char) then
        Last_Error := (Occurred => True,
                       Name     => Illegal_Literal);
      end if;
    end loop;
  end Construct_String_Literal;

  -----------------------------------------------------------------------
  -- Visible Lexer operations

  procedure Initialize (Input: in String_Pointer; Size_of_Tab: in Positive) is
    procedure Determine_Language is
      -- See above for the allowed set of initial tokens.
      First_Token: Token_Name renames Name (Get_Token);
      procedure Resolve_Package is
        -- Only identifiers and dots may follow until the deciding
        -- "is" or ';'. We also consider reserved words, because they may
        -- be identifiers in the other language.
        Next_Token: Token_Name;
      begin
        loop
          Next_Token := Name (Get_Token);
          case Next_Token is
            when Identifier_AJ |
                 Dot_AJ        => null;
            when Reserved_Word => if Next_Token = Is_A then
                                    Program_Language := Ada;
                                    return;
                                  end if;
            when Semicolon_AJ  => Program_Language := Java;
                                  return;
            when others        => raise Unidentifyable_Language;
          end case;
        end loop;
      end Resolve_Package;
    begin
      case First_Token is
        when Comment_A            |
             Function_A           |
             Generic_A            |
             Pragma_A             |
             Private_AJ           |
             Procedure_A          |
             Separate_A           |
             With_A               => Program_Language := Ada;
        when EndOfLineComment_J   |
             TraditionalComment_J |
             CommentHead_J        |
             Documentation_J      |
             DocumentationHead_J  |
             Abstract_AJ          |
             Class_J              |
             Final_J              |
             Import_J             |
             Interface_J          |
             Public_J             => Program_Language := Java;
        when Package_AJ           => Resolve_Package;
        when others               => raise Unidentifyable_Language;
      end case;
    exception
      when others => raise Unidentifyable_Language;
    end Determine_Language;
  begin
    Input_Location := Input'First;
    Input_String   := Input;
    Current_Line   := 1;
    Current_Col    := 1;
    Tab_Size := Size_of_Tab;
    Skip_White_Space;
    Program_Language    := Unknown;
    Comment_to_Complete := None;
    Determine_Language;
    -- Reset.
    Comment_to_Complete := None;
    Input_Location := Input'First;
    Current_Line   := 1;
    Current_Col    := 1;
    Skip_White_Space;
  end Initialize;

  -- Current position within Input

  function Line return Positive is
  begin
    return Current_Line;
  end Line;

  function Col return Positive is
  begin
    return Current_Col;
  end Col;

  function End_of_Input return Boolean is
  begin                                             -- Input_Location points
    return Input_Location > Input_String'Last + 1;  -- always one ahead of
  end End_of_Input;                                 -- the last character

  -----------------------------------------------------------------------

  function Get_Next_Token return Token is

    -- This is a simple LL(1) descent lexer, which checks the first
    -- character and then branches to an appropriate handler.
    -- If the first character is not sufficient to identify the token,
    -- a nested if is used.

  begin

    if Comment_to_Complete /= None then
      Input_Location := Input_Location - 1;  -- for Get_Rest_of_Comment
      Current_Col    := Current_Col - 1;
      return Complete_Java_Comment;
    end if;

    case Next_Char is
      when EOT =>
        raise End_Error;
      when '(' =>
        return Construct (Left_Parenthesis_AJ);
      when ')' =>
        return Construct (Right_Parenthesis_AJ);
      when '[' =>
        if Program_Language = Java then
          return Construct (LeftBracket_J);
        else
          return Construct_Bad_Token ((Occurred => True,
                                       Name     => Non_Language_Character),
                                      Input_Location - 1, Input_Location - 1,
                                      Current_Line, Current_Col - 1);
        end if;
      when ']' =>
        if Program_Language = Java then
          return Construct (RightBracket_J);
        else
          return Construct_Bad_Token ((Occurred => True,
                                       Name     => Non_Language_Character),
                                      Input_Location - 1, Input_Location - 1,
                                      Current_Line, Current_Col - 1);
        end if;
      when '{' =>
        if Program_Language = Java then
          return Construct (LeftBrace_J);
        else
          return Construct_Bad_Token ((Occurred => True,
                                       Name     => Non_Language_Character),
                                      Input_Location - 1, Input_Location - 1,
                                      Current_Line, Current_Col - 1);
        end if;
      when '}' =>
        if Program_Language = Java then
          return Construct (RightBrace_J);
        else
          return Construct_Bad_Token ((Occurred => True,
                                       Name     => Non_Language_Character),
                                      Input_Location - 1, Input_Location - 1,
                                      Current_Line, Current_Col - 1);
        end if;
      when '&' =>
        case Program_Language is
          when Ada  =>
            return Construct (Concatenate_A);
          when Java =>
            if Look_Ahead_is ('&') then
              return Construct (ShortCutAnd_J);
            elsif Look_Ahead_is ('=') then
              return Construct (AndAssign_J);
            else
              return Construct (And_J);
            end if;
          when Unknown =>
            null;  -- deliberately let it crash
        end case;
      when '|' =>
        case Program_Language is
          when Ada  =>
            return Construct (Alternative_A);
          when Java =>
            if Look_Ahead_is ('|') then
              return Construct (ShortCutOr_J);
            elsif Look_Ahead_is ('=') then
              return Construct (OrAssign_J);
            else
              return Construct (Or_J);
            end if;
          when Unknown =>
            null;  -- deliberately let it crash
        end case;
      when '!' =>
        if Program_Language = Java then
          if Look_Ahead_is ('=') then
            return Construct (NotEqual_J);
          else
            return Construct (Not_J);
          end if;
        else
          return Construct_Bad_Token ((Occurred => True,
                                       Name     => Non_Language_Character),
                                      Input_Location - 1, Input_Location - 1,
                                      Current_Line, Current_Col - 1);
        end if;
      when '?' =>
        if Program_Language = Java then
          return Construct (Conditional_J);
        else
          return Construct_Bad_Token ((Occurred => True,
                                       Name     => Non_Language_Character),
                                      Input_Location - 1, Input_Location - 1,
                                      Current_Line, Current_Col - 1);
        end if;
      when '%' =>
        if Program_Language = Java then
          if Look_Ahead_is ('=') then
            return Construct (RemainderAssign_J);
          else
            return Construct (Remainder_J);
          end if;
        else
          return Construct_Bad_Token ((Occurred => True,
                                       Name     => Non_Language_Character),
                                      Input_Location - 1, Input_Location - 1,
                                      Current_Line, Current_Col - 1);
        end if;
      when '^' =>
        if Program_Language = Java then
          if Look_Ahead_is ('=') then
            return Construct (XorAssign_J);
          else
            return Construct (Xor_J);
          end if;
        else
          return Construct_Bad_Token ((Occurred => True,
                                       Name     => Non_Language_Character),
                                      Input_Location - 1, Input_Location - 1,
                                      Current_Line, Current_Col - 1);
        end if;
      when '~' =>
        if Program_Language = Java then
          return Construct (Complement_J);
        else
          return Construct_Bad_Token ((Occurred => True,
                                       Name     => Non_Language_Character),
                                      Input_Location - 1, Input_Location - 1,
                                      Current_Line, Current_Col - 1);
        end if;
      when ',' =>
        return Construct (Comma_AJ);
      when ';' =>
        return Construct (Semicolon_AJ);
      when ':' =>
        if Program_Language = Ada and then Look_Ahead_is ('=') then
          return Construct (Assignment_A);
        else
          return Construct (Colon_AJ);
        end if;
      when '+' =>
        if Program_Language = Java then
          if Look_Ahead_is ('+') then
            return Construct (Increment_J);
          elsif Look_Ahead_is ('=') then
            return Construct (PlusAssign_J);
          end if;
        end if;
        return Construct (Plus_AJ);
      when '-' =>
        if Program_Language = Java then
          if Look_Ahead_is ('-') then
            return Construct (Decrement_J);
          elsif Look_Ahead_is ('=') then
            return Construct (MinusAssign_J);
          end if;
        elsif Look_Ahead_is ('-') then
          return Construct_Comment_Line;
        end if;
        return Construct (Minus_AJ);
      when '*' =>
        if Program_Language = Ada and then Look_Ahead_is ('*') then
          return Construct (Exponentiate_A);
        elsif Program_Language = Java and then Look_Ahead_is ('=') then
          return Construct (TimesAssign_J);
        else
          return Construct (Times_AJ);
        end if;
      when '/' =>
        if Look_Ahead_is ('=') then
          if Program_Language = Ada then
            return Construct (Not_Equal_A);
          else
            return Construct (DivideAssign_J);
          end if;
        elsif Program_Language /= Ada then  -- language might still be unknown
          if Look_Ahead_is ('/') then
            return Construct_Comment_Line;
          elsif Look_Ahead_is ('*') then
            return Construct_other_Java_Comment;
          end if;
        end if;
        return Construct (Divide_AJ);
      when ''' =>
        return Construct_Character_Literal_or_Tick;
      when '.' =>
        if Program_Language = Ada and then Look_Ahead_is ('.') then
          return Construct (Double_Dot_A);
        else
          return Construct (Dot_AJ);
        end if;
      when '=' =>
        case Program_Language is
          when Ada =>
            if Look_Ahead_is ('>') then
              return Construct (Arrow_A);
            else
              return Construct (Equal_A);
            end if;
          when Java =>
            if Look_Ahead_is ('=') then
              return Construct (Equal_J);
            else
              return Construct (Assignment_J);
            end if;
          when Unknown => null;  -- deliberately let it crash for unknown language
        end case;
      when '<' =>
        if Look_Ahead_is ('=') then
          return Construct (Less_Equal_AJ);
        elsif Look_Ahead_is ('<') then
          case Program_Language is
            when Ada     => return Construct (Left_Label_Bracket_A);
            when Java    => if Look_Ahead_is ('=') then
                              return Construct (LeftShiftAssign_J);
                            else
                              return Construct (LeftShift_J);
                            end if;
            when Unknown => null;  -- see above
          end case;
        elsif Program_Language = Ada and then Look_Ahead_is ('>') then
          return Construct (Box_A);
        else
          return Construct (Less_Than_AJ);
        end if;
      when '>' =>
        if Look_Ahead_is ('=') then
          return Construct (Greater_Equal_AJ);
        elsif Look_Ahead_is ('>') then
          case Program_Language is
            when Ada     => return Construct (Right_Label_Bracket_A);
            when Java    => if Look_Ahead_is ('>') then
                              if Look_Ahead_is ('=') then
                                return Construct (UnsignedRightShiftAssign_J);
                              else
                                return Construct (UnsignedRightShift_J);
                              end if;
                            elsif Look_Ahead_is ('=') then
                              return Construct (RightShiftAssign_J);
                            else
                              return Construct (RightShift_J);
                            end if;
            when Unknown => null;  -- see above
          end case;
        else
          return Construct (Greater_Than_AJ);
        end if;
      when '"' =>
        return Construct_String_Literal;
      when '_' =>  -- look ahead
        case Program_Language is
          when Ada =>
            Next_Char := Get_Next_Char;
            Put_Back_Last_Character;
            case Next_Char is
              when 'A' .. 'Z' | 'a' .. 'z' =>
                return Construct_Identifier_or_Keyword;
              when '0' .. '9' =>
                return Construct_Number;
              when others =>
                return Construct_Bad_Token ((Occurred => True,
                                             Name     => Non_Language_Character),
                                            Input_Location - 1, Input_Location - 1,
                                            Current_Line, Current_Col - 1);
            end case;
          when Java | Unknown =>
            return Construct_Identifier_or_Keyword;
        end case;
      when 'A' .. 'Z' | 'a' .. 'z' =>
        return Construct_Identifier_or_Keyword;
      when '$' =>
        case Program_Language is
          when Java    |
               Unknown => return Construct_Identifier_or_Keyword;
          when Ada     => return Construct_Bad_Token ((Occurred => True,
                                                       Name     => Non_Language_Character),
                                                      Input_Location - 1, Input_Location - 1,
                                                      Current_Line, Current_Col - 1);
        end case;
      when '0' .. '9' =>
        return Construct_Number;
      when others =>
        return Construct_Bad_Token ((Occurred => True,
                                     Name     => Non_Language_Character),
                                    Input_Location - 1, Input_Location - 1,
                                    Current_Line, Current_Col - 1);
    end case;
  end Get_Next_Token;

  -----------------------------------------------------------------------

  function Get_Token return Token is
    Next: constant Token := Get_Next_Token;  -- we are after white space
  begin
    Skip_White_Space;  -- do it after reading next token to enable test
    return Next;       -- for end of input string
  end Get_Token;

  procedure Reset (to_Token: in Token) is
  begin
    if To_Lower (Image (to_Token)) /= To_Lower (Input_String (to_Token.Core.First ..
                                                              to_Token.Core.Last  )) then
      raise Reset_Error;
    else
      Input_Location := to_Token.Core.First;
      Current_Line   := to_Token.Core.Line;
      Current_Col    := to_Token.Core.Col;
      Next_Char      := Get_Next_Char;  -- consume the first character
      case to_Token.Core.Name is
        when CommentBody_J       .. CommentTail_J       => Comment_to_Complete := Simple;
        when DocumentationBody_J .. DocumentationTail_J => Comment_to_Complete := Document;
        when others =>                                     Comment_to_Complete := None;
      end case;
    end if;
  exception
    when others => raise Reset_Error;
  end Reset;

  -----------------------------------------------------------------------

  procedure Adjust (Object: in out Token) is
    -- Deep copy.
  begin
    case Object.Core.Name is
      when Number'First ..
           Bad_Token_AJ => Object.Core.Image := new String'(Object.Core.Image.all);
      when others       => null;
    end case;
  end Adjust;

  procedure Finalize (Object: in out Token) is
  begin
    case Object.Core.Name is
      when Number'First ..
           Bad_Token_AJ => Free (Object.Core.Image);
      when others       => null;
    end case;
  end Finalize;

end Lexer;
