------------------------------------------------------------------------
-- This software is provided as is in the hope that it might be found
-- useful.
-- You may use and modify it freely provided you keep this copyright
-- notice unchanged and mark modifications appropriately.
--
-- Bug reports and proposals for improvements are welcome. Please send
-- them to the eMail address below.
--
-- Christoph Karl Walter Grein
-- Hauptstr. 42
-- D-86926 Greifenberg
-- Germany
--
-- eMail: Christ-Usch.Grein@T-Online.de
--
-- Copyright (c) 1998 Christoph Karl Walter Grein
------------------------------------------------------------------------

with Ada.Command_Line, Ada.Strings.Fixed;
with Ada.Text_IO;
use  Ada.Text_IO;

with Lexer.Read_File;  pragma Elaborate_All (Lexer);
use  Lexer;

procedure Make_Body is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.0
  -- Date      5 November 1998
  --====================================================================
  -- From a (generic) package specification, create the corresponding
  -- body.
  -- The following declarations are completed:
  -- - An incomplete type declaration;
  -- - a subprogram declaration (currently pragma Import, should it
  --   be present, is ignored);
  -- - a task (type) declaration;
  -- - a protected declaration.
  -- The formatting is taken from the specification.
  -- Precondition is a legal Ada compilation unit.
  --
  -- Call with the name of a file containing the specification as the
  -- sole command parameter.
  -- The name of the body file is the same as that of the specification
  -- with the extension replaced by the traditional "adb" (assuming an
  -- extension "ads" for the specification). If there is no extension,
  -- ".adb" is appended.
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.   0.0  25.10.1998  Sketch
  --  C.G.   0.1  30.10.1998  PDL
  --  C.G.   1.0  05.11.1998  Finished coding
  --====================================================================

  -- Vernacular:
  -- We are "at" a given token, when Get_Token will return this token;
  -- we are "after" a given token, when Get_Token will return the token
  -- following the current one.

  No_Package_Specification: exception;

  procedure Check_Input_File is

    -- Check for a sequence of tokens as
    --   [generic] package BODY Defining_Program_Unit_Name RENAMES is NEW ...
    -- The capitalised tokens must not be present.

    -- When leaving this procedure, we will be at the keyword "package"
    -- of the package specification.

    Current, Previous, Keyword: Token;

  begin

    if Ada.Command_Line.Argument_Count /= 1 then
      raise Constraint_Error;
    end if;

    Initialize (Lexer.Read_File (Ada.Command_Line.Argument (1)),
                Size_of_Tab => 2);
    -- This is a storage leak, since the file pointer is not deallocated.
    -- The program is however considered to be used stand-alone.

    Current  := Get_Token;
    Previous := Current;

    -- Loop until the keyword "package" is found which is not within a
    -- generic part.
    while not (Name (Previous) /= With_A and Name (Current) = Package_AJ) loop
      Previous := Current;
      Current  := Get_Token;
    end loop;

    -- Now we are at "package".
    Keyword := Current;

    -- It must not be followed by "body".
    if Name (Get_Token) = Body_A then
      raise No_Package_Specification;
    end if;

    -- It is a package instantiation when "is new" follows;
    -- it is a renaming when "renames" follows.
    loop
      Current := Get_Token;
      if Name (Current) = Renames_A then
        raise No_Package_Specification;
      end if;
      exit when Name (Current) = Is_A;
    end loop;
    if Name (Get_Token) = New_AJ then
      raise No_Package_Specification;
    end if;

    -- The file given as (sole) command argument is a (generic) package specification.
    Reset (Keyword);

  exception
    when Lexer.End_Error =>  -- some nonesense input was given
      raise No_Package_Specification;

  end Check_Input_File;

  -------------------------------------------------------------------------

  function Body_File_Name return String is
    Spec_File_Name: String renames Ada.Command_Line.Argument (1);
    Dot: Natural := Ada.Strings.Fixed.Index (Source  => Spec_File_Name,
                                             Pattern => ".",
                                             Going   => Ada.Strings.Backward);
  begin
    if Dot = 0 then  -- no extension
      Dot := Spec_File_Name'Last;
    else
      Dot := Dot - 1;
    end if;
    return Spec_File_Name (1 .. Dot) & ".adb";
  end Body_File_Name;

  -------------------------------------------------------------------------

  Body_File: File_Type;

  -------------------------------------------------------------------------
  -- Recursive call, thus a declaration is needed:

  procedure Create_Package_Body (Keyword: in Token);

  -------------------------------------------------------------------------
  -- When leaving these procedures, we will be after ")" it there are
  -- parameters (or discriminants), else at the place where we entered.

  procedure Skip_Parameters is
    Current : Token := Get_Token;
  begin
    if Name (Current) = Left_Parenthesis_AJ then
      loop
        Current := Get_Token;
        exit when Name (Current) = Right_Parenthesis_AJ;
      end loop;
    else
      Reset (Current);
    end if;
  end Skip_Parameters;

  procedure Copy_Parameters is
    Current : Token := Get_Token;
  begin
    if Name (Current) = Left_Parenthesis_AJ then
      loop
        Set_Col (Body_File, Count (Col (Current)));
        Put     (Body_File, Image (Current));
        exit when Name (Current) = Right_Parenthesis_AJ;
        Current := Get_Token;
      end loop;
    else
      Reset (Current);
    end if;
  end Copy_Parameters;

  -------------------------------------------------------------------------

  procedure Skip_Nested_Package is

    -- The current token is "package".
    -- When leaving this procedure, we will be after
    -- - "renames" for renaming declaration;
    -- - "new" for a generic instantiation;
    -- - "end" for a package declaration.

    Current  : Token   := Get_Token;  -- identifier
    End_Ahead: Boolean := False;

  begin

    if Name (Get_Token) = Renames_A or else Name (Get_Token) = New_AJ then
      return;
    end if;

    loop
      case Name (Get_Token) is
        when End_A =>
          exit when not End_Ahead;
          End_Ahead := False;
          Current   := Get_Token;  -- might be "end record"
        when Package_AJ =>
          Skip_Nested_Package;
        when Null_A =>  -- might be "null record", no "end" at the end
          Current := Get_Token;
        when Protected_AJ | Record_A =>  -- there will be "end" at the end
          End_Ahead := True;
        when Task_A =>  -- if it has entries, there will be "end" at the end
          if Name (Get_Token) /= Identifier_AJ then  -- it's "type"
            Current := Get_Token;  -- identifier
            Skip_Parameters;  -- discriminants here
          end if;
          End_Ahead := Name (Get_Token) = Is_A;
        when others =>
          null;
      end case;
    end loop;

  end Skip_Nested_Package;

  -------------------------------------------------------------------------

  procedure Complete_Types (First_Token: in Token) is

    -- First all incompete types (except task types) are completed.
    -- A type is incomplete when its name is followed by a semicolon.

    -- First_Token is the first token after "is" of the package
    -- specification.
    -- When leaving this procedure, we will be at First_Token again.

    Current   : Token := First_Token;
    Identifier: Token;

    End_Ahead: Boolean := False;

    Column: Ada.Text_IO.Count;

    function is_not_Completed (Identifier: Token) return Boolean is
      -- Look whether the type is completed later directly within the
      -- current package, i.e. we have to skip nested packages.
      -- Upon return be careful to reset to the token after Identifier
      -- so that we are at the same place as when entering.
      Current  : Token;
      End_Ahead: Boolean := False;
    begin
      loop
        case Name (Get_Token) is
          when End_A =>
            exit when not End_Ahead;
            End_Ahead := False;
            Current   := Get_Token;  -- might be "end record"
          when Package_AJ=>
            Skip_Nested_Package;
          when Null_A =>  -- might be "null record", no "end" at the end
            Current := Get_Token;
          when Protected_AJ | Record_A =>  -- there will be "end" at the end
            End_Ahead := True;
          when Task_A =>  -- if it has entries, there will be "end" at the end
            if Name (Get_Token) /= Identifier_AJ then  -- it's "type"
              Current := Get_Token;  -- identifier
              if Image (Current) = Image (Identifier) then  -- it's completed by a task type
                Reset (Identifier);  Current := Get_Token;  -- consume identifier
                return False;
              end if;
              Skip_Parameters;  -- discriminants here
            end if;
            End_Ahead := Name (Get_Token) = Is_A;
          when Type_A =>
            if Image (Get_Token) = Image (Identifier) then
              Reset (Identifier);  Current := Get_Token;  -- consume identifier
              return False;
            end if;
          when others =>
            null;
        end case;
      end loop;
      Reset (Identifier);  Current := Get_Token;  -- consume identifier
      return True;  -- no completion found
    end is_not_Completed;

  begin

    loop

      case Name (Current) is
      
        when Type_A =>  -- cannot be a task type (see below)

          Column     := Count (Col (Current));  -- for formatting
          Identifier := Get_Token;

          Skip_Parameters;  -- discriminants here

          if Name (Get_Token) = Semicolon_AJ and then is_not_Completed (Identifier) then
            Set_Col  (Body_File, Column);
            Put      (Body_File, "type " & Image (Identifier));
            Copy_Parameters;
            Put_Line (Body_File, " is [type_definition];");
            New_Line (Body_File);
          end if;

        when Null_A =>  -- might be "null record", no "end" at the end

          Current := Get_Token;

        when Protected_AJ | Record_A =>  -- there will be "end" at the end

          End_Ahead := True;

        when Task_A =>  -- if it has entries, there will be "end" at the end

          if Name (Get_Token) /= Identifier_AJ then  -- it's "type"
            Current := Get_Token;  -- identifier
            Skip_Parameters;       -- discriminants here
          end if;

          End_Ahead := Name (Get_Token) = Is_A;

        when Package_AJ =>

          Skip_Nested_Package;

        when End_A =>

          exit when not End_Ahead;

          End_Ahead := False;
          Current   := Get_Token;  -- might be "end record"

        when others =>

          null;

      end case;

      Current := Get_Token;

    end loop;

    Reset (First_Token);  -- go back to the starting point

  end Complete_Types;

  -------------------------------------------------------------------------

  procedure Complete_Subprogram (Keyword: in Token) is

    -- We are after the keyword "function", "procedure" or "entry".
    -- Check that it is not abstract, a generic instantiation nor a
    -- renaming.
    -- When leaving this procedure, we will be after
    -- - "is" for an abstract subprogram or a generic instantiation;
    -- - "renames" for a renaming declaration;
    -- - the terminating semicolon for a subprogram or entry declaration.

    Current   : Token  := Get_Token;  -- identifier
    Identifier: String renames Image (Current);

    procedure Skip_Return is
    begin
      Current := Get_Token;  -- "return"
      loop
        Current := Get_Token;
        exit when Name (Current) /= Identifier_AJ and Name (Current) /= Dot_AJ;
      end loop;
    end Skip_Return;

    procedure Copy_Return is
    begin
      Current := Get_Token;  Set_Col (Body_File, Count (Col (Current)));  Put (Body_File, Image (Current));
      loop
        Current := Get_Token;
        exit when Name (Current) /= Identifier_AJ and Name (Current) /= Dot_AJ;
        Set_Col (Body_File, Count (Col (Current)));  Put (Body_File, Image (Current));
      end loop;
    end Copy_Return;

  begin

    Skip_Parameters;
    if Name (Keyword) = Function_A then
      Skip_Return;
    else
      Current := Get_Token;
    end if;

    if Name (Current) = Is_A or Name (Current) = Renames_A then
      return;
    end if;

    Reset (Keyword);
    Current := Get_Token;  Set_Col (Body_File, Count (Col (Current)));  Put (Body_File, Image (Current));
    Current := Get_Token;  Set_Col (Body_File, Count (Col (Current)));  Put (Body_File, Image (Current));
    Copy_Parameters;

    case Name (Keyword) is
      when Function_A => Copy_Return;
      when Entry_A    => Put (Body_File, " when [Condition]");
      when others     => null;
    end case;

    Put (Body_File, " is"           );  Set_Col (Body_File, Count (Col (Keyword)));
    Put (Body_File, "begin"         );  Set_Col (Body_File, Count (Col (Keyword)));
    Put (Body_File, "  [statements]");  Set_Col (Body_File, Count (Col (Keyword)));
    if Name (Keyword) = Function_A then
      Put (Body_File, "  return [expression];");  Set_Col (Body_File, Count (Col (Keyword)));
    end if;
    Put_Line (Body_File, "end " & Identifier & ';');
    New_Line (Body_File);
  
  end Complete_Subprogram;

  -------------------------------------------------------------------------

  procedure Complete_Protected (Keyword: in Token) is

    -- We are after the keyword "protected".
    -- When leaving this procedure, we will be after the terminating
    -- semicolon.

    Current   : Token := Get_Token;  -- "type" or identifier
    Identifier: Token;
  
  begin

    if Name (Current) = Type_A then
      Identifier := Get_Token;
    else
      Identifier := Current;
    end if;

    Set_Col  (Body_File, Count (Col (Keyword)));
    Put_Line (Body_File, "protected body " & Image (Identifier) & " is");
    New_Line (Body_File);

    loop  -- skip until is
      exit when Name (Get_Token) = Is_A;
    end loop;

    loop
      Current := Get_Token;
      case Name (Current) is
        when Procedure_A |
             Function_A  |
             Entry_A     => Complete_Subprogram (Keyword => Current);
        when End_A       => exit;
        when others      => null;
      end case;
    end loop;

    Set_Col  (Body_File, Count (Col (Keyword)));
    Put_Line (Body_File, "end " & Image (Identifier) & ';');
    New_Line (Body_File);
  
  end Complete_Protected;

  -------------------------------------------------------------------------

  procedure Complete_Entry (Keyword: in Token) is

    -- We are after the keyword "entry".
    -- When leaving this procedure, we will be after the terminating
    -- semicolon.

    Identifier: Token := Get_Token;

  begin

    Set_Col  (Body_File, Count (Col (Keyword) + 2));
    Put      (Body_File, "when Condition =>");
    Set_Col  (Body_File, Count (Col (Keyword) + 4));
    Put      (Body_File, "accept " & Image (Identifier));
    Copy_Parameters;
    Put_Line (Body_File, " do");
    Set_Col  (Body_File, Count (Col (Keyword) + 6));
    Put      (Body_File, "[statements]");
    Set_Col  (Body_File, Count (Col (Keyword) + 4));
    Put_Line (Body_File, "end " & Image (Identifier) & ';');

  end Complete_Entry;

  -------------------------------------------------------------------------

  procedure Complete_Task (Keyword: in Token) is

    -- We are after the keyword "task".
    -- When leaving this procedure, we will be after the terminating
    -- semicolon.

    Current   : Token := Get_Token;  -- "type" or identifier
    Identifier: Token;

    Later_Entry: Boolean := False;
  
  begin

    if Name (Current) = Type_A then
      Identifier := Get_Token;
    else
      Identifier := Current;
    end if;

    Set_Col  (Body_File, Count (Col (Keyword)));
    Put_Line (Body_File, "task body " & Image (Identifier) & " is");
    Set_Col  (Body_File, Count (Col (Keyword)));
    Put_Line (Body_File, "begin");

    loop  -- skip until "is" or ";"
      Current := Get_Token;
      case Name (Current) is
        when Semicolon_AJ |
             Is_A         => exit;
        when others       => null;
      end case;
    end loop;

    if Name (Current) = Is_A then

      Set_Col  (Body_File, Count (Col (Keyword)));
      Put_Line (Body_File, "  select");

      loop
        Current := Get_Token;
        case Name (Current) is
          when Entry_A => if Later_Entry then
                            Set_Col  (Body_File, Count (Col (Keyword)));
                            Put_Line (Body_File, "  or");
                          end if;
                          Later_Entry := True;
                          Complete_Entry (Keyword => Current);
          when End_A   => exit;
          when others  => null;
        end case;
      end loop;

      Set_Col  (Body_File, Count (Col (Keyword)));
      Put_Line (Body_File, "  end select;");

    else

      Set_Col  (Body_File, Count (Col (Keyword)));
      Put_Line (Body_File, "  [statements]");

    end if;

    Set_Col  (Body_File, Count (Col (Keyword)));
    Put_Line (Body_File, "end " & Image (Identifier) & ';');
    New_Line (Body_File);

  end Complete_Task;

  -------------------------------------------------------------------------

  procedure Complete_Program_Units is

    -- Program units are completed in the order they appear in the
    -- specification.

    -- We are at the first token after "is" of the package
    -- specification.
    -- When leaving this procedure, we will be at the semicolon at the
    -- package end.

    Current: Token := Get_Token;

    End_Ahead: Boolean := False;

  begin

    loop

      case Name (Current) is

        when Function_A | Procedure_A =>

          Complete_Subprogram (Keyword => Current);

        when Package_AJ =>  -- must not be a generic instantiation nor a renaming

          declare
            Keyword   : constant Token := Current;
            Identifier: constant Token := Get_Token;
            Is_Renames: constant Token := Get_Token;  -- "is" or "renames"
          begin
            if Name (Is_Renames) = Is_A and then Name (Get_Token) /= New_AJ then
              Reset (Keyword);
              Create_Package_Body (Keyword => Get_Token);
            end if;
          end;

        when Protected_AJ =>

          Complete_Protected (Keyword => Current);

        when Task_A =>

          Complete_Task (Keyword => Current);

        when Null_A   |   -- might be "null record", no "end" at the end
             Access_A =>  -- might be "access procedure" or "access function"

          Current := Get_Token;

        when Record_A =>  -- there will be "end" at the end

          End_Ahead := True;

        when End_A =>

          exit when not End_Ahead;

          End_Ahead := False;
          Current   := Get_Token;  -- "end record"

        when others =>

          null;

      end case;

      Current := Get_Token;

    end loop;

  end Complete_Program_Units;

  -------------------------------------------------------------------------

  procedure Create_Package_Body (Keyword: in Token) is

    -- We are after the keyword "package" of a package for which a body has
    -- to be created.
    -- When leaving this procedure, we will be at the semicolon at the
    -- end of the package specification.

    Identifier: constant Token := Get_Token;
    Current   :          Token := Identifier;
    Previous  , Last:    Token;

    Column: constant Ada.Text_IO.Count := Count (Col (Keyword));

    procedure Write_Defining_Program_Unit_Name (Identifier: in Token) is
      -- For the outermost level, the Defining_Program_Unit_Name can be
      -- a Parent_Unit_Name or the Defining_Identifier.
      -- When leaving this procedure, we will be after "is".
      Current: Token;
    begin
      Reset (Identifier);
      loop
        Current := Get_Token;
        exit when Name (Current) = Is_A;
        Put (Body_File, Image (Current));
      end loop;
    end Write_Defining_Program_Unit_Name;

  begin

    -- Write "package body Identifier is".

    Set_Col  (Body_File, Column);
    Put      (Body_File, "package body ");
    Write_Defining_Program_Unit_Name (Identifier);
    Put_Line (Body_File," is");
    New_Line (Body_File);

    -- Complete constructs

    Complete_Types (First_Token => Get_Token);
    Complete_Program_Units;

    -- Write "end Identifier;".

    Last := Get_Token;  -- remember last position

    Set_Col  (Body_File, Column);
    Put      (Body_File, "end ");
    Write_Defining_Program_Unit_Name (Identifier);
    Put_Line (Body_File, ";");
    New_Line (Body_File);

    Reset (Last);

  end Create_Package_Body;

begin  --------------------------------------------------------------------

  Check_Input_File;  -- Now the next token returned from Lexer is the identifier.

  Create (Name => Body_File_Name, File => Body_File);

  Create_Package_Body (Keyword => Get_Token);

  Close (Body_File);

exception
  when No_Package_Specification =>
    Put_Line (Ada.Text_IO.Standard_Error,
              "No (generic) package specification given.");
  when Ada.Text_IO.Name_Error =>
    Put_Line (Ada.Text_IO.Standard_Error,
              "File " & Ada.Command_Line.Argument (1) & " not found.");
  when others =>
    Put_Line (Ada.Text_IO.Standard_Error,
              "Call with the name of a file specifying a (generic) package.");

end Make_Body;
