------------------------------------------------------------------------
-- This software is provided as is in the hope that it might be found
-- useful.
-- You may use and modify it freely provided you keep this copyright
-- notice unchanged and mark modifications appropriately.
--
-- Bug reports and proposals for improvements are welcome. Please send
-- them to the eMail address below.
--
-- Christoph Karl Walter Grein
-- Hauptstr. 42
-- D-86926 Greifenberg
-- Germany
--
-- eMail: Christ-Usch.Grein@T-Online.de
--
-- Copyright (c) 1998 Christoph Karl Walter Grein
------------------------------------------------------------------------

with Ada.Text_IO;
use  Ada.Text_IO;
with Ada.Characters.Latin_1;

with Ada.Strings.Unbounded;

with Lexer;  -- Testee
pragma Elaborate_All (Lexer);

procedure Test_Lexer is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   2.1
  -- Date      5 August 1998
  --====================================================================
  -- Test program for Lexer.
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  23.06.1998
  --  C.G.    1.1  01.07.1998 Amend test for EOL (Windows95 and Unix)
  --  C.G.    1.2  06.07.1998 Size of horizontal tabulation
  --  C.G.    2.0  29.07.1998 Java
  --  C.G.    2.1  05.08.1998 Added Documentation_Tag
  --====================================================================

  CR: Character renames Ada.Characters.Latin_1.CR;   -- carriage return
  LF: Character renames Ada.Characters.Latin_1.LF;   -- line feed
  HT: Character renames Ada.Characters.Latin_1.HT;   -- horizontal tabulation

  function "+" (Source: String) return Ada.Strings.Unbounded.Unbounded_String renames
    Ada.Strings.Unbounded.To_Unbounded_String;
  function "+" (Source: Ada.Strings.Unbounded.Unbounded_String) return String renames
    Ada.Strings.Unbounded.To_String;

  type Result (Name : Lexer.Token_Name := Lexer.Bad_Token_AJ) is record
    Image: Ada.Strings.Unbounded.Unbounded_String;
    case Name is
      when Lexer.Based_Integer_AJ   |
           Lexer.BasedLongInteger_J |
           Lexer.Based_Real_A       => Base : Ada.Text_IO.Number_Base;
      when Lexer.Bad_Token_AJ       => Error: Lexer.Token_Error;
      when Lexer.Documentation      => Tag  : Lexer.Documentation_Tag;
      when others                   => null;
    end case;
  end record;
  type Expectation is array (Positive range <>) of Result;

  Plural     : constant array (Boolean) of Lexer.String_Pointer :=
                (False => new String'(""),
                 True  => new String'("s"));
  Correctness: constant array (Boolean) of Lexer.String_Pointer :=
                (False => new String'("Error"),
                 True  => new String'("OK"));

  Ada_Language : aliased constant String := "with ";   -- initial tokens to
  Java_Language: aliased constant String := "class ";  -- set language

  Number_of_Errors, Number_of_Keyword_Errors,
  Number_of_Ada_Errors, Number_of_Java_Errors: Natural := 0;

  procedure Test (Testee: in String; Language: in String; Expected: in Expectation) is
    Number: Natural := Expected'First - 1;
    Token: Lexer.Token;
    use type Lexer.Token_Name, Lexer.Token_Error, Lexer.Documentation_Tag;
  begin
    for C in Testee'Range loop
      if Testee (C) = CR then
        Put ("<CR>");
      elsif Testee (C) = LF then
        Put ("<LF>");
      elsif Testee (C) = HT then
        Put ("<HT>");
      else
        Put (Testee (C));
      end if;
    end loop;
    New_Line;
    Lexer.Initialize (new String'(Language & Testee), Size_of_Tab => 8);
    Token := Lexer.Get_Token;  -- ignore language deciding first token
    while not Lexer.End_of_Input loop
      Token  := Lexer.Get_Token;
      Number := Number + 1;
      begin
        if ( Lexer.Name  (Token) /=  Expected (Number).Name ) or else
           ( Lexer.Image (Token) /= +Expected (Number).Image) or else
           ((Lexer.Name  (Token)  = Lexer.Based_Integer_AJ    or
             Lexer.Name  (Token)  = Lexer.BasedLongInteger_J  or
             Lexer.Name  (Token)  = Lexer.Based_Real_A )      and then 
            (Lexer.Base  (Token) /=  Expected (Number).Base)) or else
           ( Lexer.Name  (Token)  =  Lexer.Bad_Token_AJ       and then
             Lexer.Error (Token) /=  Expected (Number).Error) or else
           ( Lexer.Name  (Token) in Lexer.Documentation       and then
             Lexer.Tag   (Token) /=  Expected (Number).Tag  ) then
          Number_of_Errors := Number_of_Errors + 1;
          if Language = Ada_Language then
            Number_of_Ada_Errors := Number_of_Ada_Errors + 1;
          else
            Number_of_Java_Errors := Number_of_Java_Errors + 1;
          end if;
          Put_Line ("Erroneous result");
          Put_Line ("Number  : " & Integer'Image (Number));
          Put_Line ("Expected: " & Lexer.Token_Name'Image (Expected (Number).Name));
          Put_Line ("Found   : " & Lexer.Token_Name'Image (Lexer.Name (Token)));
          Put_Line ("Image   : " & Lexer.Image (Token));
          if (Lexer.Name (Token)  = Lexer.Based_Integer_AJ   or
              Lexer.Name  (Token) = Lexer.BasedLongInteger_J or
              Lexer.Name  (Token) = Lexer.Based_Real_A )     then
            Put_Line ("Base    : " & Integer'Image (Lexer.Base (Token)));
          elsif Lexer.Name (Token) in Lexer.Documentation then
            Put_Line ("Tag     : " & Lexer.Documentation_Tag'Image (Lexer.Tag (Token)));
          elsif Lexer.Name (Token) = Lexer.Bad_Token_AJ then
            Put_Line ("Error   : " & Lexer.Token_Error'Image (Lexer.Error (Token)));
          end if;
        end if;
      exception
        when Constraint_Error =>
          Number_of_Errors := Number_of_Errors + 1;
          if Language = Ada_Language then
            Number_of_Ada_Errors := Number_of_Ada_Errors + 1;
          else
            Number_of_Java_Errors := Number_of_Java_Errors + 1;
          end if;
          Put_Line ("Erroneous result");
          Put_Line ("Number  : " & Integer'Image (Number));
          Put_Line ("Expected: none");
          Put_Line ("Found   : " & Lexer.Token_Name'Image (Lexer.Name (Token)));
          Put_Line ("Image   : " & Lexer.Image (Token));
          if Lexer.Name (Token) = Lexer.Bad_Token_AJ then
            Put_Line ("Error   : " & Lexer.Token_Error'Image (Lexer.Error (Token)));
          end if;
      end;
    end loop;
    Put_Line (Integer'Image (Number) & " token" &
              Plural (Number /= 1).all & " found (" &
              Correctness (Number = Expected'Last).all & ')');
    New_Line;  Flush;
  end Test;

  procedure Test_Keywords is separate;
  procedure Test_Ada      is separate;
  procedure Test_Java     is separate;

begin

  Test_Keywords;
  Test_Ada;
  Test_Java;

  -- Result

  Ada.Text_IO.Put_Line ("Test result:");
  Ada.Text_IO.Put_Line (Integer'Image (Number_of_Errors)         & " errors found.");
  Ada.Text_IO.Put_Line (Integer'Image (Number_of_Keyword_Errors) & " keyword errors found.");
  Ada.Text_IO.Put_Line (Integer'Image (Number_of_Ada_Errors)     & " Ada errors found.");
  Ada.Text_IO.Put_Line (Integer'Image (Number_of_Java_Errors)    & " Java errors found.");

end Test_Lexer;
